import copy
import os
import shutil
import collections
import datetime
import json
import heapq
from typing import Any, Dict, Optional, Union, List, Tuple, Set, Callable
import proxai.types as types
import proxai.serializers.type_serializer as type_serializer
import proxai.serializers.hash_serializer as hash_serializer
import proxai.state_controllers.state_controller as state_controller

CACHE_DIR = 'query_cache'
LIGHT_CACHE_RECORDS_PATH = 'light_cache_records.json'
_QUERY_CACHE_MANAGER_STATE_PROPERTY = '_query_cache_manager_state'


def _to_light_cache_record(cache_record: types.CacheRecord):
  return types.LightCacheRecord(
      query_record_hash=cache_record.query_record.hash_value,
      query_response_count=len(cache_record.query_responses),
      shard_id=cache_record.shard_id,
      last_access_time=cache_record.last_access_time,
      call_count=cache_record.call_count)


def _get_cache_size(
    cache_record: Union[types.CacheRecord, types.LightCacheRecord]) -> int:
  if isinstance(cache_record, types.LightCacheRecord):
    return cache_record.query_response_count + 1
  return len(cache_record.query_responses) + 1


def _get_hash_value(
    cache_record: Union[
        str,
        types.CacheRecord,
        types.LightCacheRecord,
        types.QueryRecord]
) -> str:
  if isinstance(cache_record, str):
    return cache_record
  if isinstance(cache_record, types.CacheRecord):
    if cache_record.query_record.hash_value:
      return cache_record.query_record.hash_value
    else:
      cache_record.query_record.hash_value = (
          hash_serializer.get_query_record_hash(
              cache_record.query_record))
      return cache_record.query_record.hash_value
  if isinstance(cache_record, types.LightCacheRecord):
    if cache_record.query_record_hash:
      return cache_record.query_record_hash
    else:
      raise ValueError('LightCacheRecord doesn\'t have query_record_hash')
  if isinstance(cache_record, types.QueryRecord):
    query_record = cache_record
    if query_record.hash_value:
      return query_record.hash_value
    else:
      query_record.hash_value = hash_serializer.get_query_record_hash(
          query_record)
      return query_record.hash_value


class HeapManager:
  _heap: List[Tuple[int, Union[int, str]]]
  _active_values: Dict[Union[int, str], int]
  _record_size_map: Dict[Union[int, str], int]
  _with_size: bool
  _total_size: int

  def __init__(self, with_size: bool = False):
    self._heap = []
    self._active_values = {}
    self._record_size_map = {}
    self._with_size = with_size
    if with_size:
      self._total_size = 0

  def push(
      self,
      key: Union[int, str],
      value: int,
      record_size: int = None):
    if not self._with_size and record_size:
      raise ValueError('Cannot push record size without with_size=True')
    if self._with_size and not record_size:
      raise ValueError('Cannot push without record size with with_size=False')
    if key in self._active_values and self._with_size:
      self._total_size -= self._record_size_map[key]
    self._active_values[key] = value
    if self._with_size:
      self._record_size_map[key] = record_size
      self._total_size += record_size
    heapq.heappush(self._heap, (value, key))

  def pop(self) -> Optional[Tuple[int, Union[int, str]]]:
    while self._heap:
      value, key = heapq.heappop(self._heap)
      if key in self._active_values and self._active_values[key] == value:
        del self._active_values[key]
        if self._with_size:
          self._total_size -= self._record_size_map[key]
          del self._record_size_map[key]
        return value, key
    return None, None

  def top(self) -> Optional[Tuple[int, Union[int, str]]]:
    while self._heap:
      value, key = self._heap[0]
      if key in self._active_values and self._active_values[key] == value:
        return value, key
      heapq.heappop(self._heap)
    return None, None

  def __len__(self):
    if self._with_size:
      return self._total_size
    return len(self._active_values)


class ShardManager:
  _path: str
  _shard_count: int
  _response_per_file: int
  _shard_paths: Dict[Union[int, str], str]
  _backlog_shard_path: str
  _light_cache_records_path: str
  _shard_active_count: Dict[Union[int, str], int]
  _shard_heap: HeapManager
  _loaded_cache_records: Dict[str, types.CacheRecord]
  _light_cache_records: Dict[str, types.LightCacheRecord]
  _map_shard_to_cache: Dict[Union[str, int], Set[str]]

  def __init__(
      self,
      path: str,
      shard_count: int,
      response_per_file: int):
    if shard_count < 1 or shard_count > 99999:
      raise ValueError('shard_count should be between 1 and 99999')
    if response_per_file < 1:
      raise ValueError('response_per_file should be greater than 0')

    self._path = path
    self._shard_count = shard_count
    self._response_per_file = response_per_file
    self._shard_paths = {}
    self._backlog_shard_path = None
    self._light_cache_records_path = None
    self._load_light_cache_records()

  @property
  def backlog_shard_path(self):
    if self._backlog_shard_path:
      return self._backlog_shard_path
    self._backlog_shard_path = os.path.join(
        self._path, f'shard_backlog_{self._shard_count:05}.jsonl')
    return self._backlog_shard_path

  @property
  def shard_paths(self):
    if self._shard_paths:
      return self._shard_paths
    self._shard_paths = {}
    for i in range(self._shard_count):
      self._shard_paths[i] = os.path.join(
          self._path, f'shard_{i:05}-of-{self._shard_count:05}.jsonl')
    self._shard_paths['backlog'] = self.backlog_shard_path
    return self._shard_paths

  @property
  def light_cache_records_path(self) -> str:
    if self._light_cache_records_path:
      return self._light_cache_records_path
    self._light_cache_records_path = os.path.join(
        self._path, f'light_cache_records_{self._shard_count:05}.json')
    return self._light_cache_records_path

  def _check_shard_id(self, shard_id: Union[int, str]):
    if shard_id not in self.shard_paths:
      raise ValueError('Invalid shard_id')

  def _update_cache_record(
      self,
      cache_record: Union[types.CacheRecord, types.LightCacheRecord],
      delete_only: bool = False,
      write_to_file: bool = True):
    hash_value = _get_hash_value(cache_record)
    shard_id = cache_record.shard_id
    if isinstance(cache_record, types.CacheRecord):
      light_cache_record = _to_light_cache_record(cache_record)
    else:
      light_cache_record = cache_record

    # Clean previous values
    if hash_value in self._light_cache_records:
      old_shard_id = self._light_cache_records[hash_value].shard_id
      self._map_shard_to_cache[old_shard_id].remove(hash_value)
      self._shard_active_count[old_shard_id] -= _get_cache_size(
          self._light_cache_records[hash_value])
      if old_shard_id != 'backlog':
        self._shard_heap.push(
            key=old_shard_id, value=self._shard_active_count[old_shard_id])
      if hash_value in self._loaded_cache_records:
        del self._loaded_cache_records[hash_value]
      del self._light_cache_records[hash_value]
      if write_to_file:
        with open(self._light_cache_records_path, 'a') as f:
          f.write(json.dumps({hash_value: {}}))
          f.write('\n')
    if delete_only:
      return

    # Insert new values
    self._light_cache_records[hash_value] = light_cache_record
    self._map_shard_to_cache[shard_id].add(hash_value)
    self._shard_active_count[shard_id] += _get_cache_size(
        light_cache_record)
    if shard_id != 'backlog':
      self._shard_heap.push(
          key=shard_id, value=self._shard_active_count[shard_id])
    if isinstance(cache_record, types.CacheRecord):
      self._loaded_cache_records[hash_value] = cache_record
    if write_to_file:
      with open(self._light_cache_records_path, 'a') as f:
        f.write(json.dumps(
            {hash_value: type_serializer.encode_light_cache_record(
              light_cache_record)}))
        f.write('\n')

  def _save_light_cache_records(self):
    data = {}
    for query_record_hash, light_cache_record in (
        self._light_cache_records.items()):
      data[query_record_hash] = type_serializer.encode_light_cache_record(
          light_cache_record)
    try:
      os.rename(
          self._light_cache_records_path,
          self._light_cache_records_path + '_backup')
    except OSError:
      pass

    with open(self._light_cache_records_path, 'w') as f:
      f.write(json.dumps(data))
      f.write('\n')

  def _load_light_cache_records(self):
    # Reset all values
    self._shard_active_count = {}
    self._shard_heap = HeapManager()
    self._loaded_cache_records = {}
    self._light_cache_records = {}
    self._map_shard_to_cache = collections.defaultdict(set)
    for shard_id in self.shard_paths:
      self._shard_active_count[shard_id] = 0
      if shard_id != 'backlog':
        self._shard_heap.push(key=shard_id, value=0)
    data = {}

    # Load light cache records from backup if primary file is corrupted
    def load_data(file_path: str):
      data = {}
      with open(file_path, 'r') as f:
        for line in f:
          try:
            record_data = json.loads(line)
          except Exception:
            continue
          for hash_value, record in record_data.items():
            data[hash_value] = record
      return data
    try:
      data = load_data(self.light_cache_records_path)
    except Exception as e1:
      try:
        data = load_data(self.light_cache_records_path + '_backup')
      except Exception as e2:
        return

    # Load light cache records from data
    for query_record_hash, record in data.items():
      if record == {}:
        continue
      try:
        light_cache_record = type_serializer.decode_light_cache_record(record)
      except Exception:
        continue
      if query_record_hash != light_cache_record.query_record_hash:
        continue
      if (isinstance(light_cache_record.shard_id, str)
          and light_cache_record.shard_id != 'backlog'):
        continue
      if (isinstance(light_cache_record.shard_id, int)
          and (light_cache_record.shard_id < 0
            or light_cache_record.shard_id >= self._shard_count)):
        continue
      light_cache_record.call_count = 0
      self._update_cache_record(light_cache_record, write_to_file=False)
    self._save_light_cache_records()

  def _check_cache_record_is_up_to_date(
      self, cache_record: types.CacheRecord) -> bool:
    hash_value = _get_hash_value(cache_record)
    if hash_value not in self._light_cache_records:
      return False
    light_cache_record = copy.deepcopy(
        self._light_cache_records[hash_value])
    comparison_light_cache_record = copy.deepcopy(
        _to_light_cache_record(
            cache_record))
    light_cache_record.call_count = None
    comparison_light_cache_record.call_count = None
    if light_cache_record != comparison_light_cache_record:
      return False
    return True

  def _load_shard(
      self, shard_id: Union[int, str]) ->  List[str]:
    result = []
    try:
      with open(self.shard_paths[shard_id], 'r') as f:
        for line in f:
          try:
            cache_record = type_serializer.decode_cache_record(json.loads(line))
          except Exception:
            continue
          if not self._check_cache_record_is_up_to_date(cache_record):
            continue
          cache_record.call_count = self._light_cache_records[
              cache_record.query_record.hash_value].call_count
          self._update_cache_record(cache_record)
          result.append(_get_hash_value(cache_record))
    except Exception:
      pass
    for hash_value in list(self._map_shard_to_cache[shard_id]):
      if hash_value not in result:
        light_cache_value = copy.deepcopy(
            self._light_cache_records[hash_value])
        self._update_cache_record(light_cache_value, delete_only=True)
    return result

  def _move_backlog_to_shard(self, shard_id: Union[int, str]):
    self._check_shard_id(shard_id)
    if shard_id == 'backlog':
      raise ValueError('Cannot move backlog to backlog')
    shard_hash_values = self._load_shard(shard_id=shard_id)
    backlog_hash_values = self._load_shard(shard_id='backlog')
    for hash_value in backlog_hash_values:
      cache_record = copy.deepcopy(self._loaded_cache_records[hash_value])
      cache_record.shard_id = shard_id
      self._update_cache_record(cache_record)

    with open(self.shard_paths[shard_id] + '_backup', 'w') as f:
      for hash_value in shard_hash_values + backlog_hash_values:
        try:
          cache_record = self._loaded_cache_records[hash_value]
          f.write(json.dumps(type_serializer.encode_cache_record(cache_record)))
          f.write('\n')
        except Exception:
          continue
    os.rename(
        self.shard_paths[shard_id] + '_backup',
        self.shard_paths[shard_id])
    try:
      os.remove(self.shard_paths['backlog'])
    except OSError:
      pass

  def _add_to_backlog(self, cache_record: types.CacheRecord):
    cache_record = copy.deepcopy(cache_record)
    cache_record.shard_id = 'backlog'
    self._update_cache_record(cache_record)
    with open(self.shard_paths['backlog'], 'a') as f:
      f.write(json.dumps(type_serializer.encode_cache_record(cache_record)))
      f.write('\n')

  def get_cache_record(
      self, query_record: Union[types.QueryRecord, str]) -> Optional[types.CacheRecord]:
    hash_value = _get_hash_value(query_record)
    if hash_value not in self._light_cache_records:
      return None
    light_cache_record = self._light_cache_records[hash_value]
    if light_cache_record.shard_id not in self.shard_paths:
      return None
    self._load_shard(shard_id=light_cache_record.shard_id)
    if hash_value not in self._loaded_cache_records:
      return None
    return self._loaded_cache_records[hash_value]

  def delete_record(
      self,
      cache_record: Union[
          str,
          types.CacheRecord,
          types.LightCacheRecord,
          types.QueryRecord]):
    hash_value = _get_hash_value(cache_record)
    if hash_value not in self._light_cache_records:
      return
    light_cache_records = copy.deepcopy(
        self._light_cache_records[hash_value])
    self._update_cache_record(light_cache_records, delete_only=True)

  def save_record(self, cache_record: types.CacheRecord):
    hash_value = _get_hash_value(cache_record)
    self.delete_record(hash_value)

    backlog_size = self._shard_active_count['backlog']
    record_size = _get_cache_size(cache_record)
    lowest_shard_value, lowest_shard_key = self._shard_heap.top()
    if (backlog_size + record_size
        > self._response_per_file - lowest_shard_value):
      self._move_backlog_to_shard(shard_id=lowest_shard_key)
      self._add_to_backlog(cache_record)
      self._save_light_cache_records()
    else:
      self._add_to_backlog(cache_record)


class QueryCacheManager(state_controller.StateControlled):
  _cache_options: types.CacheOptions
  _shard_count: int
  _response_per_file: int
  _cache_response_size: int
  _query_cache_manager_state: types.QueryCacheManagerState
  _shard_manager: ShardManager
  _record_heap: HeapManager

  def __init__(
      self,
      cache_options: Optional[types.CacheOptions] = None,
      get_cache_options: Optional[Callable[[], types.CacheOptions]] = None,
      shard_count: Optional[int] = None,
      response_per_file: Optional[int] = None,
      cache_response_size: Optional[int] = None,
      init_state: Optional[types.QueryCacheManagerState] = None):

    if init_state and (
        cache_options is not None or
        get_cache_options is not None or
        shard_count is not None or
        response_per_file is not None or
        cache_response_size is not None):
      raise ValueError(
          'init_state and other parameters cannot be set at the same time.')

    super().__init__(
        cache_options=cache_options,
        get_cache_options=get_cache_options,
        shard_count=shard_count,
        response_per_file=response_per_file,
        cache_response_size=cache_response_size)

    self.init_state()
    self.set_property_value(
        'status', types.QueryCacheManagerStatus.INITIALIZING)

    if init_state:
      self.load_state(init_state)
      self._init_dir()
      self._init_managers()
    else:
      initial_state = self.get_state()
      self._get_cache_options = get_cache_options
      self.cache_options = cache_options
      if response_per_file is not None:
        self.response_per_file = response_per_file
      if shard_count is not None:
        self.shard_count = shard_count
      if cache_response_size is not None:
        self.cache_response_size = cache_response_size
      self.handle_changes(initial_state, self.get_state())

  def get_internal_state_property_name(self):
    return _QUERY_CACHE_MANAGER_STATE_PROPERTY

  def get_internal_state_type(self):
    return types.QueryCacheManagerState

  def handle_changes(
      self,
      old_state: types.QueryCacheManagerState,
      current_state: types.QueryCacheManagerState):
    if current_state.cache_options is None:
      self.status = types.QueryCacheManagerStatus.CACHE_OPTIONS_NOT_FOUND
      return

    if current_state.cache_options.cache_path is None:
      self.status = types.QueryCacheManagerStatus.CACHE_PATH_NOT_FOUND
      return

    if not os.access(current_state.cache_options.cache_path, os.W_OK):
      self.status = types.QueryCacheManagerStatus.CACHE_PATH_NOT_WRITABLE
      return

    if (old_state.cache_options is None or
        old_state.cache_options.cache_path !=
        current_state.cache_options.cache_path):
      self._init_dir()

    if (
        old_state.cache_options != current_state.cache_options or
        old_state.shard_count != current_state.shard_count or
        old_state.response_per_file != current_state.response_per_file or
        old_state.cache_response_size != current_state.cache_response_size):
      self._init_managers()

    self.status = types.QueryCacheManagerStatus.WORKING

  def clear_cache(self):
    if (
        self.status == types.QueryCacheManagerStatus.INITIALIZING or
        self.status == types.QueryCacheManagerStatus.CACHE_OPTIONS_NOT_FOUND or
        self.status == types.QueryCacheManagerStatus.CACHE_PATH_NOT_FOUND or
        self.status == types.QueryCacheManagerStatus.CACHE_PATH_NOT_WRITABLE):
      raise ValueError(f'QueryCacheManager status is {self.status}')
    cache_dir = self._get_cache_dir(self.cache_options.cache_path)
    if os.path.exists(cache_dir):
      shutil.rmtree(cache_dir)
    self._init_dir()
    self._init_managers()

  def _init_dir(self):
    if self.cache_options.cache_path is None:
      return
    os.makedirs(
        self._get_cache_dir(self.cache_options.cache_path),
        exist_ok=True)

  def _init_managers(self):
    if self.cache_options.cache_path is None:
      return
    self._shard_manager = ShardManager(
        path=self._get_cache_dir(self.cache_options.cache_path),
        shard_count=self.shard_count,
        response_per_file=self.response_per_file)
    self._record_heap = HeapManager(with_size=True)
    for record in self._shard_manager._light_cache_records.values():
      self._push_record_heap(record)

  @staticmethod
  def _get_cache_dir(cache_path: str) -> str:
    return os.path.join(cache_path, CACHE_DIR)

  @property
  def status(self) -> types.QueryCacheManagerStatus:
    return self.get_property_value('status')

  @status.setter
  def status(self, value: types.QueryCacheManagerStatus):
    self.set_property_value('status', value)

  @property
  def cache_options(self) -> types.CacheOptions:
    return self.get_property_value('cache_options')

  @cache_options.setter
  def cache_options(self, value: types.CacheOptions):
    self.set_property_value('cache_options', value)

  @property
  def shard_count(self) -> int:
    return self.get_property_value('shard_count')

  @shard_count.setter
  def shard_count(self, value: int):
    self.set_property_value('shard_count', value)

  @property
  def response_per_file(self) -> int:
    return self.get_property_value('response_per_file')

  @response_per_file.setter
  def response_per_file(self, value: int):
    self.set_property_value('response_per_file', value)

  @property
  def cache_response_size(self) -> int:
    return self.get_property_value('cache_response_size')

  @cache_response_size.setter
  def cache_response_size(self, value: int):
    self.set_property_value('cache_response_size', value)

  def _push_record_heap(
      self, cache_record: Union[types.CacheRecord, types.LightCacheRecord]):
    hash_value = _get_hash_value(cache_record)
    last_access_time = cache_record.last_access_time.timestamp()
    self._record_heap.push(
        key=hash_value,
        value=last_access_time,
        record_size=_get_cache_size(cache_record))
    while len(self._record_heap) > self.cache_response_size:
      _, hash_value = self._record_heap.pop()
      self._shard_manager.delete_record(hash_value)

  def look(
      self,
      query_record: types.QueryRecord,
      update: bool = True,
      unique_response_limit: Optional[int] = None,
  ) -> types.CacheLookResult:
    if self.status != types.QueryCacheManagerStatus.WORKING:
      raise ValueError(f'QueryCacheManager status is {self.status}')

    if not isinstance(query_record, types.QueryRecord):
      raise ValueError('query_record should be of type QueryRecord')
    cache_record = self._shard_manager.get_cache_record(query_record)
    if cache_record is None:
      return types.CacheLookResult(
          look_fail_reason=types.CacheLookFailReason.CACHE_NOT_FOUND)
    if cache_record.query_record != query_record:
      return types.CacheLookResult(
          look_fail_reason=types.CacheLookFailReason.CACHE_NOT_MATCHED)
    if unique_response_limit == None:
      unique_response_limit = self.cache_options.unique_response_limit
    if len(cache_record.query_responses) < unique_response_limit:
      return types.CacheLookResult(
          look_fail_reason=
          types.CacheLookFailReason.UNIQUE_RESPONSE_LIMIT_NOT_REACHED)
    query_response: types.QueryResponseRecord = cache_record.query_responses[
        cache_record.call_count % len(cache_record.query_responses)]
    if (query_response.error
        and self.cache_options.retry_if_error_cached
        and cache_record.call_count < len(cache_record.query_responses)):
      cache_record.last_access_time = datetime.datetime.now()
      cache_record.call_count += 1
      self._shard_manager.save_record(cache_record=cache_record)
      self._push_record_heap(cache_record)
      return types.CacheLookResult(
          look_fail_reason=types.CacheLookFailReason.PROVIDER_ERROR_CACHED)
    if update:
      cache_record.last_access_time = datetime.datetime.now()
      cache_record.call_count += 1
      self._shard_manager.save_record(cache_record=cache_record)
      self._push_record_heap(cache_record)
    return types.CacheLookResult(query_response=query_response)

  def cache(
      self,
      query_record: types.QueryRecord,
      response_record: types.QueryResponseRecord,
      unique_response_limit: Optional[int] = None):
    if self.status != types.QueryCacheManagerStatus.WORKING:
      raise ValueError(f'QueryCacheManager status is {self.status}')

    current_time = datetime.datetime.now()
    cache_record = self._shard_manager.get_cache_record(query_record)
    if not cache_record:
      query_record.hash_value = hash_serializer.get_query_record_hash(
          query_record)
      cache_record = types.CacheRecord(
          query_record=query_record,
          query_responses=[response_record],
          last_access_time=current_time,
          call_count=0)
      self._shard_manager.save_record(cache_record=cache_record)
      self._push_record_heap(cache_record)
      return
    if unique_response_limit == None:
      unique_response_limit = self.cache_options.unique_response_limit
    if len(cache_record.query_responses) < unique_response_limit:
      cache_record.query_responses.append(response_record)
      cache_record.last_access_time = current_time
      self._shard_manager.save_record(cache_record=cache_record)
      self._push_record_heap(cache_record)
      return
    if (self.cache_options.retry_if_error_cached
        and response_record.error == None):
      for idx, previous_response in enumerate(cache_record.query_responses):
        if previous_response.error:
          cache_record.query_responses[idx] = response_record
          cache_record.last_access_time = current_time
          self._shard_manager.save_record(cache_record=cache_record)
          self._push_record_heap(cache_record)
          return



from typing import Optional


class _MockResponse(object):
  text: str

  def __init__(self):
    self.text = 'mock response'


class _MockModel(object):
  def generate_content(
      self,
      model,
      config,
      contents) -> _MockResponse:
    return _MockResponse()


class GeminiMock(object):
  models: _MockModel

  def __init__(self):
    self.models = _MockModel()



import os
import copy
import json
import requests
from functools import wraps
import proxai.types as types
import proxai.experiment.experiment as experiment
import proxai.logging.utils as logging_utils
import proxai.state_controllers.state_controller as state_controller
from typing import Any, Callable, Dict, List, Optional, Union, Tuple

_PROXDASH_STATE_PROPERTY = '_proxdash_connection_state'
_NOT_SET_EXPERIMENT_PATH_VALUE = '(not set)'


class ProxDashConnection(state_controller.StateControlled):
  _status: Optional[types.ProxDashConnectionStatus]
  _hidden_run_key: Optional[str]
  _experiment_path: Optional[str]
  _get_experiment_path: Optional[Callable[[], str]]
  _logging_options: Optional[types.LoggingOptions]
  _get_logging_options: Optional[Callable[[], types.LoggingOptions]]
  _proxdash_options: Optional[types.ProxDashOptions]
  _get_proxdash_options: Optional[Callable[[], types.ProxDashOptions]]
  _key_info_from_proxdash: Optional[Dict]
  _connected_experiment_path: Optional[str]
  _proxdash_connection_state: Optional[types.ProxDashConnectionState]

  def __init__(
      self,
      hidden_run_key: Optional[str] = None,
      experiment_path: Optional[str] = None,
      get_experiment_path: Optional[Callable[[], str]] = None,
      logging_options: Optional[types.LoggingOptions] = None,
      get_logging_options: Optional[Callable[[], types.LoggingOptions]] = None,
      proxdash_options: Optional[types.ProxDashOptions] = None,
      get_proxdash_options: Optional[
          Callable[[], types.ProxDashOptions]] = None,
      init_state: Optional[types.ProxDashConnectionState] = None):
    if init_state and (
        hidden_run_key is not None or
        experiment_path is not None or get_experiment_path is not None or
        logging_options is not None or get_logging_options is not None or
        proxdash_options is not None or get_proxdash_options is not None):
      raise ValueError(
          'If init_state is provided, none of the other arguments should be '
          'provided.')

    super().__init__(
        hidden_run_key=hidden_run_key,
        experiment_path=experiment_path,
        get_experiment_path=get_experiment_path,
        logging_options=logging_options,
        get_logging_options=get_logging_options,
        proxdash_options=proxdash_options,
        get_proxdash_options=get_proxdash_options)

    self.init_state()
    self.set_property_value(
        'status', types.ProxDashConnectionStatus.INITIALIZING)

    if init_state:
      self.load_state(init_state)
    else:
      initial_state = self.get_state()
      self._get_experiment_path = get_experiment_path
      self._get_logging_options = get_logging_options
      self._get_proxdash_options = get_proxdash_options

      self.hidden_run_key = hidden_run_key
      self.logging_options = logging_options
      self.proxdash_options = proxdash_options
      self.experiment_path = experiment_path
      self.handle_changes(initial_state, self.get_state())

  def get_internal_state_property_name(self):
    return _PROXDASH_STATE_PROPERTY

  def get_internal_state_type(cls):
    return types.ProxDashConnectionState

  def handle_changes(
      self,
      old_state: types.ProxDashConnectionState,
      current_state: types.ProxDashConnectionState):
    result_state = copy.deepcopy(old_state)
    if current_state.logging_options is not None:
      result_state.logging_options = current_state.logging_options
    if current_state.proxdash_options is not None:
      result_state.proxdash_options = current_state.proxdash_options
    if current_state.key_info_from_proxdash is not None:
      result_state.key_info_from_proxdash = current_state.key_info_from_proxdash
    if current_state.experiment_path is not None:
      result_state.experiment_path = current_state.experiment_path
    if current_state.connected_experiment_path is not None:
      result_state.connected_experiment_path = (
          current_state.connected_experiment_path)
    if current_state.status is not None:
      result_state.status = current_state.status

    if result_state.proxdash_options is None:
      raise ValueError(
          'ProxDash options are not set for both old and new states. '
          'This creates an invalid state change.')
    if result_state.logging_options is None:
      raise ValueError(
          'Logging options are not set for both old and new states. '
          'This creates an invalid state change.')

    proxdash_disabled = result_state.proxdash_options.disable_proxdash
    if proxdash_disabled:
      self.status = types.ProxDashConnectionStatus.DISABLED
      self.key_info_from_proxdash = None
      # Note: There is no longer any connection to ProxDash. This change
      # shouldn't be logged, so, self.connected_experiment_path setter should
      # not be used here.
      self.set_property_value_without_triggering_getters(
          'connected_experiment_path', None)
      return

    if result_state.proxdash_options.api_key is None:
      if 'PROXDASH_API_KEY' not in os.environ:
        self.status = types.ProxDashConnectionStatus.API_KEY_NOT_FOUND
        self.key_info_from_proxdash = None
        # Note: There is no longer any connection to ProxDash. This change
        # shouldn't be logged, so, self.connected_experiment_path setter should
        # not be used here.
        self.set_property_value_without_triggering_getters(
            'connected_experiment_path', None)
        return
      else:
        # Note: Setting api_key from environment variable.
        self.proxdash_options.api_key = os.environ['PROXDASH_API_KEY']
        current_state.proxdash_options.api_key = os.environ['PROXDASH_API_KEY']
        result_state.proxdash_options.api_key = os.environ['PROXDASH_API_KEY']

    api_key_query_required = False

    old_api_key = None
    if old_state.proxdash_options:
      old_api_key = old_state.proxdash_options.api_key
    if old_api_key != result_state.proxdash_options.api_key:
      api_key_query_required = True
    if old_api_key == result_state.proxdash_options.api_key and (
        result_state.status == types.ProxDashConnectionStatus.INITIALIZING or
        result_state.status == types.ProxDashConnectionStatus.DISABLED or
        result_state.status == types.ProxDashConnectionStatus.API_KEY_NOT_FOUND
    ):
      api_key_query_required = True

    old_base_url = types.ProxDashOptions.base_url
    if old_state.proxdash_options:
      old_base_url = old_state.proxdash_options.base_url
    if old_base_url != result_state.proxdash_options.base_url:
      api_key_query_required = True

    if api_key_query_required:
      validation_status, key_info_from_proxdash = self._check_api_key_validity(
          base_url=result_state.proxdash_options.base_url,
          api_key=result_state.proxdash_options.api_key)
      result_state.status = validation_status
      result_state.key_info_from_proxdash = key_info_from_proxdash

    if result_state.status == types.ProxDashConnectionStatus.API_KEY_NOT_VALID:
      self.status = types.ProxDashConnectionStatus.API_KEY_NOT_VALID
      self.key_info_from_proxdash = None
      raise ValueError(
          'ProxDash API key not valid. Please provide a valid API key.\n'
          f'base_url: {result_state.proxdash_options.base_url}\n'
          f'api_key: {result_state.proxdash_options.api_key[:3]}...\n\n'
          'To fix this issue:\n'
          '1. Check that your PROXDASH_API_KEY in your .bashrc or .zshrc file '
          'is correct if it exists\n'
          '2. Check that px.ProxDashOptions(api_key="your_api_key") is correct '
          'if you set it directly\n'
          '3. Verify that your key matches what appears on '
          'https://proxai.co/dashboard/api-keys\n'
          '4. If you don\'t want to use ProxDash, make sure PROXDASH_API_KEY '
          'is not set in your environment variables\n'
          'For more information, see: '
          'https://www.proxai.co/proxai-docs/advanced/proxdash-connection')

    if result_state.status == types.ProxDashConnectionStatus.PROXDASH_INVALID_RETURN:
      self.status = types.ProxDashConnectionStatus.PROXDASH_INVALID_RETURN
      self.key_info_from_proxdash = None
      raise ValueError(
          'ProxDash returned an invalid response.\nPlease report this '
          'issue to the https://github.com/proxai/proxai.\n'
          'Also, please check latest stable version of ProxAI.')

    if result_state.status != types.ProxDashConnectionStatus.CONNECTED:
      raise ValueError(
          'Unknown ProxDash connection status.\n'
          f'result_state.status: {result_state.status}\n'
          'result_state.key_info_from_proxdash: '
          f'{result_state.key_info_from_proxdash}')

    if (self.status != result_state.status or
        old_base_url != result_state.proxdash_options.base_url or
        old_api_key != result_state.proxdash_options.api_key):
      # Note: This is good for logging purposes.
      self.status = result_state.status
    if self.key_info_from_proxdash != result_state.key_info_from_proxdash:
      self.key_info_from_proxdash = result_state.key_info_from_proxdash
    if self.connected_experiment_path != self.experiment_path:
      self.connected_experiment_path = self.experiment_path

  @property
  def hidden_run_key(self) -> Optional[str]:
    return self.get_property_value('hidden_run_key')

  @hidden_run_key.setter
  def hidden_run_key(self, hidden_run_key: Optional[str]):
    self.set_property_value('hidden_run_key', hidden_run_key)

  @property
  def logging_options(self) -> types.LoggingOptions:
    return self.get_property_value('logging_options')

  @logging_options.setter
  def logging_options(self, logging_options: types.LoggingOptions):
    self.set_property_value('logging_options', logging_options)

  @property
  def proxdash_options(self) -> types.ProxDashOptions:
    return self.get_property_value('proxdash_options')

  @proxdash_options.setter
  def proxdash_options(self, proxdash_options: types.ProxDashOptions):
    self.set_property_value('proxdash_options', proxdash_options)

  @property
  def key_info_from_proxdash(self) -> Optional[Dict]:
    return self.get_property_value('key_info_from_proxdash')

  @key_info_from_proxdash.setter
  def key_info_from_proxdash(self, key_info_from_proxdash: Optional[Dict]):
    self.set_property_value('key_info_from_proxdash', key_info_from_proxdash)

  @property
  def experiment_path(self) -> str:
    internal_experiment_path = self.get_property_internal_value(
        'experiment_path')
    internal_get_experiment_path = self.get_property_func_getter(
        'experiment_path')

    experiment_path = None
    if (internal_experiment_path is not None and
        internal_experiment_path != _NOT_SET_EXPERIMENT_PATH_VALUE):
      experiment_path = internal_experiment_path
    elif internal_get_experiment_path is not None:
      experiment_path = internal_get_experiment_path()

    if experiment_path is None:
      experiment_path = _NOT_SET_EXPERIMENT_PATH_VALUE

    self.set_property_internal_state_value(
        'experiment_path', experiment_path)
    return experiment_path

  @experiment_path.setter
  def experiment_path(self, experiment_path: Optional[str]):
    if experiment_path is not None:
      experiment.validate_experiment_path(experiment_path)
    else:
      experiment_path = _NOT_SET_EXPERIMENT_PATH_VALUE

    self.set_property_value('experiment_path', experiment_path)

  @property
  def connected_experiment_path(self) -> str:
    return self.get_property_value('connected_experiment_path')

  @connected_experiment_path.setter
  def connected_experiment_path(self, connected_experiment_path: Optional[str]):
    if self.status != types.ProxDashConnectionStatus.CONNECTED:
      if connected_experiment_path is not None:
        raise ValueError(
            'Connected experiment path can only be set if the ProxDash '
            'connection is connected.')
      self.set_property_value_without_triggering_getters(
          'connected_experiment_path', None)
      return

    previous_experiment_path = self.get_property_internal_value(
        'connected_experiment_path')
    if previous_experiment_path is None:
      previous_experiment_path = _NOT_SET_EXPERIMENT_PATH_VALUE

    new_experiment_path = connected_experiment_path
    if new_experiment_path is None:
      new_experiment_path = _NOT_SET_EXPERIMENT_PATH_VALUE

    if previous_experiment_path != new_experiment_path:
      logging_utils.log_proxdash_message(
          logging_options=self.logging_options,
          proxdash_options=self.proxdash_options,
          message=(
              'Connected to ProxDash experiment: '
              f'{new_experiment_path}'),
          type=types.LoggingType.INFO)

    self.set_property_value(
        'connected_experiment_path', connected_experiment_path)

  @property
  def status(self) -> types.ProxDashConnectionStatus:
    return self.get_property_value('status')

  @status.setter
  def status(self, status: types.ProxDashConnectionStatus):
    self.set_property_value('status', status)
    if status == types.ProxDashConnectionStatus.INITIALIZING:
      logging_utils.log_proxdash_message(
          logging_options=self.logging_options,
          proxdash_options=self.proxdash_options,
          message='ProxDash connection initializing.',
          type=types.LoggingType.INFO)
    elif status == types.ProxDashConnectionStatus.DISABLED:
      logging_utils.log_proxdash_message(
          logging_options=self.logging_options,
          proxdash_options=self.proxdash_options,
          message='ProxDash connection disabled.',
          type=types.LoggingType.INFO)
    elif status == types.ProxDashConnectionStatus.API_KEY_NOT_FOUND:
      logging_utils.log_proxdash_message(
          logging_options=self.logging_options,
          proxdash_options=self.proxdash_options,
          message=(
              'ProxDash connection disabled. Please provide a valid API key '
              'either as an argument or as an environment variable.'),
          type=types.LoggingType.ERROR)
    elif status == types.ProxDashConnectionStatus.API_KEY_NOT_VALID:
      logging_utils.log_proxdash_message(
          logging_options=self.logging_options,
          proxdash_options=self.proxdash_options,
          message=(
              'ProxDash API key not valid. Please provide a valid API key.\n'
              'Check proxai.co/dashboard/api-keys page to get your API '
              'key.'),
          type=types.LoggingType.ERROR)
    elif status == types.ProxDashConnectionStatus.PROXDASH_INVALID_RETURN:
      logging_utils.log_proxdash_message(
          logging_options=self.logging_options,
          proxdash_options=self.proxdash_options,
          message=(
              'ProxDash returned an invalid response.\nPlease report this '
              'issue to the https://github.com/proxai/proxai.\n'
              'Also, please check latest stable version of ProxAI.'),
          type=types.LoggingType.ERROR)
    elif status == types.ProxDashConnectionStatus.CONNECTED:
      logging_utils.log_proxdash_message(
          logging_options=self.logging_options,
          proxdash_options=self.proxdash_options,
          message=(
              f'Connected to ProxDash at {self.proxdash_options.base_url}'),
          type=types.LoggingType.INFO)

  def _check_api_key_validity(
      self,
      base_url: str,
      api_key: str) -> Tuple[
      Union[
          types.ProxDashConnectionStatus.API_KEY_NOT_VALID,
          types.ProxDashConnectionStatus.PROXDASH_INVALID_RETURN,
          types.ProxDashConnectionStatus.CONNECTED,
      ],
      Optional[Dict]]:
    response = requests.post(
        f'{base_url}/connect',
        data={'apiKey': api_key})
    if response.status_code != 201 or response.text == 'false':
      return types.ProxDashConnectionStatus.API_KEY_NOT_VALID, None
    try:
      key_info_from_proxdash = json.loads(response.text)
      return types.ProxDashConnectionStatus.CONNECTED, key_info_from_proxdash
    except Exception:
      return types.ProxDashConnectionStatus.PROXDASH_INVALID_RETURN, None

  def _hide_sensitive_content_logging_record(
      self, logging_record: types.LoggingRecord) -> types.LoggingRecord:
    logging_record = copy.deepcopy(logging_record)
    if logging_record.query_record and logging_record.query_record.prompt:
      logging_record.query_record.prompt = '<sensitive content hidden>'
    if logging_record.query_record and logging_record.query_record.system:
      logging_record.query_record.system = '<sensitive content hidden>'
    if logging_record.query_record and logging_record.query_record.messages:
      logging_record.query_record.messages = [
        {
          'role': 'assistant',
          'content': '<sensitive content hidden>'
        }
      ]
    if (logging_record.response_record and
        logging_record.response_record.response):
      logging_record.response_record.response = '<sensitive content hidden>'
    return logging_record

  def upload_logging_record(self, logging_record: types.LoggingRecord):
    self.apply_external_state_changes()

    if self.status != types.ProxDashConnectionStatus.CONNECTED:
      return
    if ((self.proxdash_options and
         self.proxdash_options.hide_sensitive_content) or
        self._key_info_from_proxdash['permission'] == 'NO_PROMPT'):
      logging_record = self._hide_sensitive_content_logging_record(
        logging_record)

    if logging_record.query_record.messages is not None:
      messages = json.dumps(
          logging_record.query_record.messages,
          indent=2,
          sort_keys=True)
    else:
      messages = None

    if logging_record.query_record.stop is not None:
      stop = logging_record.query_record.stop
      if type(stop) == str:
        stop = [stop]
      stop = json.dumps(stop, indent=2, sort_keys=True)
    else:
      stop = None

    data = {
      'apiKey': self.proxdash_options.api_key,
      'hiddenRunKey': self.hidden_run_key,
      'experimentPath': self.experiment_path,
      'callType': logging_record.query_record.call_type,
      'provider': logging_record.query_record.provider_model.provider,
      'model': logging_record.query_record.provider_model.model,
      'providerModelIdentifier': (
          logging_record.query_record.provider_model.provider_model_identifier),
      'prompt': logging_record.query_record.prompt,
      'system': logging_record.query_record.system,
      'messages': messages,
      'maxTokens': logging_record.query_record.max_tokens,
      'temperature': logging_record.query_record.temperature,
      'stop': stop,
      'hashValue': logging_record.query_record.hash_value,
      'queryTokens': logging_record.query_record.token_count,
      'response': logging_record.response_record.response,
      'error': logging_record.response_record.error,
      'errorTraceback': logging_record.response_record.error_traceback,
      'startUTCDate': logging_record.response_record.start_utc_date.isoformat(),
      'endUTCDate': logging_record.response_record.end_utc_date.isoformat(),
      'localTimeOffsetMinute': (
          logging_record.response_record.local_time_offset_minute),
      'responseTime': (
          logging_record.response_record.response_time.total_seconds() * 1000),
      'estimatedCost': logging_record.response_record.estimated_cost,
      'responseTokens': logging_record.response_record.token_count,
      'responseSource': logging_record.response_source,
      'lookFailReason': logging_record.look_fail_reason,
    }
    try:
      response = requests.post(
          f'{self.proxdash_options.base_url}/logging-record', json=data)
      if response.status_code != 201 or response.text != 'success':
        logging_utils.log_proxdash_message(
            logging_options=self.logging_options,
            proxdash_options=self.proxdash_options,
            message=(
                'ProxDash could not log the record. Error message:\n'
                f'{response.text}'),
            type=types.LoggingType.ERROR)
    except Exception as e:
      logging_utils.log_proxdash_message(
          logging_options=self.logging_options,
          proxdash_options=self.proxdash_options,
          message=(
              'ProxDash could not log the record. Error message:\n'
              f'{e}'),
          type=types.LoggingType.ERROR)



# read version from installed package
from importlib.metadata import version
from proxai.proxai import (
    CacheOptions,
    LoggingOptions,
    ProxDashOptions,
    set_run_type,
    check_health,
    connect,
    set_model,
    generate_text,
    get_summary,
    get_available_models,
    get_current_options,
    reset_state,
    reset_platform_cache,
    _init_globals,
)


__version__ = version("proxai")
_init_globals()
models = get_available_models()



import copy
import functools
import os
import requests
from typing import Dict, List, Optional
import proxai.types as types
import proxai.connectors.providers.huggingface_mock as huggingface_mock
import proxai.connectors.model_connector as model_connector

_MODEL_URL_MAP = {
    'Qwen/Qwen3-32B': 'https://router.huggingface.co/hf-inference/models/Qwen/Qwen2.5-Coder-32B-Instruct/v1/chat/completions',
    'deepseek-ai/DeepSeek-R1': 'https://router.huggingface.co/together/v1/chat/completions',
    'deepseek-ai/DeepSeek-V3': 'https://router.huggingface.co/together/v1/chat/completions',
    'google/gemma-2-2b-it': 'https://router.huggingface.co/nebius/v1/chat/completions',
    'meta-llama/Meta-Llama-3.1-8B-Instruct': 'https://router.huggingface.co/hf-inference/models/meta-llama/Meta-Llama-3.1-8B-Instruct/v1/chat/completions',
    'microsoft/phi-4': 'https://router.huggingface.co/nebius/v1/chat/completions'
}


class _HuggingFaceRequest:
  def __init__(self):
    self.headers = {
        'Authorization': f'Bearer {os.environ["HUGGINGFACE_API_KEY"]}'}

  def generate_content(
      self,
      messages: List[Dict[str, str]],
      model: str,
      max_tokens: Optional[int]=None,
      temperature: Optional[float]=None,
      stop: Optional[List[str]]=None) -> str:
    payload = {
        'model': model,
        'messages': messages
    }
    if max_tokens is not None:
      payload['max_tokens'] = max_tokens
    if temperature is not None:
      payload['temperature'] = temperature
    if stop is not None:
      payload['stop'] = stop
    response = requests.post(
        _MODEL_URL_MAP[model],
        headers=self.headers,
        json=payload)
    response_text = response.json()['choices'][0]['message']['content']
    return response_text


class HuggingFaceConnector(model_connector.ProviderModelConnector):
  def get_provider_name(self):
    return 'huggingface'

  def init_model(self):
    return _HuggingFaceRequest()

  def init_mock_model(self):
    return huggingface_mock.HuggingFaceMock()

  def generate_text_proc(self, query_record: types.QueryRecord) -> str:
    provider_model = query_record.provider_model
    query_messages = []
    if query_record.system is not None:
      query_messages.append({'role': 'system', 'content': query_record.system})
    if query_record.prompt is not None:
      query_messages.append({'role': 'user', 'content': query_record.prompt})
    if query_record.messages is not None:
      query_messages.extend(query_record.messages)

    create = functools.partial(
        self.api.generate_content,
        model=provider_model.provider_model_identifier,
        messages=query_messages)
    if query_record.max_tokens is not None:
      create = functools.partial(create, max_tokens=query_record.max_tokens)
    if query_record.temperature is not None:
      create = functools.partial(create, temperature=query_record.temperature)
    if query_record.stop is not None:
      if isinstance(query_record.stop, str):
        create = functools.partial(create, stop=[query_record.stop])
      else:
        create = functools.partial(create, stop=query_record.stop)
    completion = create()
    return completion



from types import MappingProxyType
from typing import Dict, Tuple, List
import proxai.types as types
import proxai.type_utils as type_utils

PROVIDER_KEY_MAP: Dict[str, Tuple[str]] = MappingProxyType({
    'claude': tuple(['ANTHROPIC_API_KEY']),
    'cohere': tuple(['CO_API_KEY']),
    'databricks': tuple(['DATABRICKS_TOKEN', 'DATABRICKS_HOST']),
    'deepseek': tuple(['DEEPSEEK_API_KEY']),
    'gemini': tuple(['GEMINI_API_KEY']),
    'grok': tuple(['XAI_API_KEY']),
    'huggingface': tuple(['HUGGINGFACE_API_KEY']),
    'mistral': tuple(['MISTRAL_API_KEY']),
    'openai': tuple(['OPENAI_API_KEY']),

    'mock_provider': tuple(['MOCK_PROVIDER_API_KEY']),
    'mock_failing_provider': tuple(['MOCK_FAILING_PROVIDER']),
    'mock_slow_provider': tuple(['MOCK_SLOW_PROVIDER']),
})


ALL_MODELS: Dict[str, Dict[str, types.ProviderModelType]] = MappingProxyType({
  # Mock provider
  'mock_provider': MappingProxyType({
    'mock_model': types.ProviderModelType(
      provider='mock_provider',
      model='mock_model',
      provider_model_identifier='mock_model'
    ),
  }),

  # Mock failing provider
  'mock_failing_provider': MappingProxyType({
    'mock_failing_model': types.ProviderModelType(
      provider='mock_failing_provider',
      model='mock_failing_model',
      provider_model_identifier='mock_failing_model'
    ),
  }),

  # Mock slow provider
  'mock_slow_provider': MappingProxyType({
    'mock_slow_model': types.ProviderModelType(
      provider='mock_slow_provider',
      model='mock_slow_model',
      provider_model_identifier='mock_slow_model'
    ),
  }),

  # OpenAI models.
  # Models provided by OpenAI:
  # https://platform.openai.com/docs/guides/text-generation
  'openai': MappingProxyType({
    'gpt-4.1': types.ProviderModelType(
      provider='openai',
      model='gpt-4.1',
      provider_model_identifier='gpt-4.1-2025-04-14'
    ),
    'gpt-4.1-mini': types.ProviderModelType(
      provider='openai',
      model='gpt-4.1-mini',
      provider_model_identifier='gpt-4.1-mini-2025-04-14'
    ),
    'gpt-4.1-nano': types.ProviderModelType(
      provider='openai',
      model='gpt-4.1-nano',
      provider_model_identifier='gpt-4.1-nano-2025-04-14'
    ),
    'gpt-4.5-preview': types.ProviderModelType(
      provider='openai',
      model='gpt-4.5-preview',
      provider_model_identifier='gpt-4.5-preview-2025-02-27'
    ),
    'gpt-4o': types.ProviderModelType(
      provider='openai',
      model='gpt-4o',
      provider_model_identifier='gpt-4o-2024-08-06'
    ),
    'gpt-4o-audio-preview': types.ProviderModelType(
      provider='openai',
      model='gpt-4o-audio-preview',
      provider_model_identifier='gpt-4o-audio-preview-2024-12-17'
    ),
    'gpt-4o-realtime-preview': types.ProviderModelType(
      provider='openai',
      model='gpt-4o-realtime-preview',
      provider_model_identifier='gpt-4o-realtime-preview-2024-12-17'
    ),
    'gpt-4o-mini': types.ProviderModelType(
      provider='openai',
      model='gpt-4o-mini',
      provider_model_identifier='gpt-4o-mini-2024-07-18'
    ),
    'gpt-4o-mini-audio-preview': types.ProviderModelType(
      provider='openai',
      model='gpt-4o-mini-audio-preview',
      provider_model_identifier='gpt-4o-mini-audio-preview-2024-12-17'
    ),
    'gpt-4o-mini-realtime-preview': types.ProviderModelType(
      provider='openai',
      model='gpt-4o-mini-realtime-preview',
      provider_model_identifier='gpt-4o-mini-realtime-preview-2024-12-17'
    ),
    'o1': types.ProviderModelType(
      provider='openai',
      model='o1',
      provider_model_identifier='o1-2024-12-17'
    ),
    'o1-pro': types.ProviderModelType(
      provider='openai',
      model='o1-pro',
      provider_model_identifier='o1-pro-2025-03-19'
    ),
    'o3': types.ProviderModelType(
      provider='openai',
      model='o3',
      provider_model_identifier='o3-2025-04-16'
    ),
    'o4-mini': types.ProviderModelType(
      provider='openai',
      model='o4-mini',
      provider_model_identifier='o4-mini-2025-04-16'
    ),
    'o3-mini': types.ProviderModelType(
      provider='openai',
      model='o3-mini',
      provider_model_identifier='o3-mini-2025-01-31'
    ),
    'o1-mini': types.ProviderModelType(
      provider='openai',
      model='o1-mini',
      provider_model_identifier='o1-mini-2024-09-12'
    ),
    'gpt-4o-mini-search-preview': types.ProviderModelType(
      provider='openai',
      model='gpt-4o-mini-search-preview',
      provider_model_identifier='gpt-4o-mini-search-preview-2025-03-11'
    ),
    'gpt-4o-search-preview': types.ProviderModelType(
      provider='openai',
      model='gpt-4o-search-preview',
      provider_model_identifier='gpt-4o-search-preview-2025-03-11'
    ),
    'computer-use-preview': types.ProviderModelType(
      provider='openai',
      model='computer-use-preview',
      provider_model_identifier='computer-use-preview-2025-03-11'
    ),
    'chatgpt-4o-latest': types.ProviderModelType(
      provider='openai',
      model='chatgpt-4o-latest',
      provider_model_identifier='chatgpt-4o-latest'
    ),
    'gpt-4-turbo': types.ProviderModelType(
      provider='openai',
      model='gpt-4-turbo',
      provider_model_identifier='gpt-4-turbo-2024-04-09'
    ),
    'gpt-4': types.ProviderModelType(
      provider='openai',
      model='gpt-4',
      provider_model_identifier='gpt-4-0613'
    ),
    'gpt-4-32k': types.ProviderModelType(
      provider='openai',
      model='gpt-4-32k',
      provider_model_identifier='gpt-4-32k'
    ),
    'gpt-3.5-turbo': types.ProviderModelType(
      provider='openai',
      model='gpt-3.5-turbo',
      provider_model_identifier='gpt-3.5-turbo-0125'
    ),
  }),

  # Claude models.
  # Models provided by Claude:
  # https://claude.ai/docs/models
  'claude': MappingProxyType({
    'sonnet':  types.ProviderModelType(
      provider='claude',
      model='sonnet',
      provider_model_identifier='claude-3-7-sonnet-20250219'
    ),
    'haiku':  types.ProviderModelType(
      provider='claude',
      model='haiku',
      provider_model_identifier='claude-3-5-haiku-20241022'
    ),
    '3.5-sonnet-v2': types.ProviderModelType(
      provider='claude',
      model='3.5-sonnet-v2',
      provider_model_identifier='claude-3-5-sonnet-20241022'
    ),
    '3.5-sonnet': types.ProviderModelType(
      provider='claude',
      model='3.5-sonnet',
      provider_model_identifier='claude-3-5-sonnet-20240620'
    ),
    'opus':  types.ProviderModelType(
      provider='claude',
      model='opus',
      provider_model_identifier='claude-3-opus-20240229'
    ),
    '3-sonnet': types.ProviderModelType(
      provider='claude',
      model='3-sonnet',
      provider_model_identifier='claude-3-sonnet-20240229'
    ),
    '3-haiku': types.ProviderModelType(
      provider='claude',
      model='3-haiku',
      provider_model_identifier='claude-3-haiku-20240307'
    ),
  }),

  # Gemini models.
  # Models provided by Gemini:
  # https://ai.google.dev/models/gemini
  'gemini': MappingProxyType({
    'gemini-1.5-pro': types.ProviderModelType(
      provider='gemini',
      model='gemini-1.5-pro',
      provider_model_identifier='gemini-1.5-pro'
    ),
    'gemini-1.5-flash-8b': types.ProviderModelType(
      provider='gemini',
      model='gemini-1.5-flash-8b',
      provider_model_identifier='gemini-1.5-flash-8b'
    ),
    'gemini-1.5-flash': types.ProviderModelType(
      provider='gemini',
      model='gemini-1.5-flash',
      provider_model_identifier='gemini-1.5-flash'
    ),
    'gemini-2.0-flash-lite': types.ProviderModelType(
      provider='gemini',
      model='gemini-2.0-flash-lite',
      provider_model_identifier='gemini-2.0-flash-lite'
    ),
    'gemini-2.0-flash': types.ProviderModelType(
      provider='gemini',
      model='gemini-2.0-flash',
      provider_model_identifier='gemini-2.0-flash'
    ),
    'gemini-2.5-flash-preview': types.ProviderModelType(
      provider='gemini',
      model='gemini-2.5-flash-preview',
      provider_model_identifier='gemini-2.5-flash-preview-05-20'
    ),
    'gemini-2.5-pro-preview': types.ProviderModelType(
      provider='gemini',
      model='gemini-2.5-pro-preview',
      provider_model_identifier='gemini-2.5-pro-preview-05-06'
    ),
  }),

  # Cohere models.
  # Models provided by Cohere:
  # https://docs.cohere.com/docs/models
  'cohere': MappingProxyType({
    'command-a': types.ProviderModelType(
      provider='cohere',
      model='command-a',
      provider_model_identifier='command-a-03-2025'
    ),
    'command-r7b': types.ProviderModelType(
      provider='cohere',
      model='command-r7b',
      provider_model_identifier='command-r7b-12-2024'
    ),
    'command-r-plus': types.ProviderModelType(
      provider='cohere',
      model='command-r-plus',
      provider_model_identifier='command-r-plus'
    ),
    'command-r-08-2024': types.ProviderModelType(
      provider='cohere',
      model='command-r-08-2024',
      provider_model_identifier='command-r-08-2024'
    ),
    'command-r': types.ProviderModelType(
      provider='cohere',
      model='command-r',
      provider_model_identifier='command-r'
    ),
    'command': types.ProviderModelType(
      provider='cohere',
      model='command',
      provider_model_identifier='command'
    ),
    'command-nightly': types.ProviderModelType(
      provider='cohere',
      model='command-nightly',
      provider_model_identifier='command-nightly'
    ),
    'command-light': types.ProviderModelType(
      provider='cohere',
      model='command-light',
      provider_model_identifier='command-light'
    ),
    'command-light-nightly': types.ProviderModelType(
      provider='cohere',
      model='command-light-nightly',
      provider_model_identifier='command-light-nightly'
    ),
  }),

  # Databricks models.
  # Models provided by Databricks:
  # https://docs.databricks.com/en/machine-learning/foundation-models/index.html#provisioned-throughput-foundation-model-apis
  'databricks': MappingProxyType({
    'llama-4-maverick': types.ProviderModelType(
      provider='databricks',
      model='llama-4-maverick',
      provider_model_identifier='databricks-llama-4-maverick'
    ),
    'claude-3-7-sonnet': types.ProviderModelType(
      provider='databricks',
      model='claude-3-7-sonnet',
      provider_model_identifier='databricks-claude-3-7-sonnet'
    ),
    'meta-llama-3-1-8b-it': types.ProviderModelType(
      provider='databricks',
      model='meta-llama-3-1-8b-it',
      provider_model_identifier='databricks-meta-llama-3-1-8b-instruct'
    ),
    'meta-llama-3-3-70b-it': types.ProviderModelType(
      provider='databricks',
      model='meta-llama-3-3-70b-it',
      provider_model_identifier='databricks-meta-llama-3-3-70b-instruct'
    ),
    # TODO: This is extremely slow model. Until better filtering, it is not
    # included in the list.
    # 'meta-llama-3-1-405b-it': types.ProviderModelType(
    #   provider='databricks',
    #   model='meta-llama-3-1-405b-it',
    #   provider_model_identifier='databricks-meta-llama-3-1-405b-instruct'
    # ),
    'dbrx-it': types.ProviderModelType(
      provider='databricks',
      model='dbrx-it',
      provider_model_identifier='databricks-dbrx-instruct'
    ),
    'mixtral-8x7b-it': types.ProviderModelType(
      provider='databricks',
      model='mixtral-8x7b-it',
      provider_model_identifier='databricks-mixtral-8x7b-instruct'
    ),
  }),

  # Mistral models.
  # Models provided by Mistral:
  # https://docs.mistral.ai/platform/endpoints/
  'mistral': MappingProxyType({
    'codestral': types.ProviderModelType(
      provider='mistral',
      model='codestral',
      provider_model_identifier='codestral-2501'
    ),
    'ministral-3b': types.ProviderModelType(
      provider='mistral',
      model='ministral-3b',
      provider_model_identifier='ministral-3b-2410'
    ),
    'ministral-8b': types.ProviderModelType(
      provider='mistral',
      model='ministral-8b',
      provider_model_identifier='ministral-8b-2410'
    ),
    'mistral-large': types.ProviderModelType(
      provider='mistral',
      model='mistral-large',
      provider_model_identifier='mistral-large-2411'
    ),
    'mistral-medium': types.ProviderModelType(
      provider='mistral',
      model='mistral-medium',
      provider_model_identifier='mistral-medium-2505'
    ),
    'mistral-saba': types.ProviderModelType(
      provider='mistral',
      model='mistral-saba',
      provider_model_identifier='mistral-saba-2502'
    ),
    'mistral-small': types.ProviderModelType(
      provider='mistral',
      model='mistral-small',
      provider_model_identifier='mistral-small-2503'
    ),
    'open-mistral-7b': types.ProviderModelType(
      provider='mistral',
      model='open-mistral-7b',
      provider_model_identifier='open-mistral-7b'
    ),
    'open-mistral-nemo': types.ProviderModelType(
      provider='mistral',
      model='open-mistral-nemo',
      provider_model_identifier='open-mistral-nemo'
    ),
    'open-mixtral-8x22b': types.ProviderModelType(
      provider='mistral',
      model='open-mixtral-8x22b',
      provider_model_identifier='open-mixtral-8x22b'
    ),
    'open-mixtral-8x7b': types.ProviderModelType(
      provider='mistral',
      model='open-mixtral-8x7b',
      provider_model_identifier='open-mixtral-8x7b'
    ),
    'pixtral-12b': types.ProviderModelType(
      provider='mistral',
      model='pixtral-12b',
      provider_model_identifier='pixtral-12b-2409'
    ),
    'pixtral-large': types.ProviderModelType(
      provider='mistral',
      model='pixtral-large',
      provider_model_identifier='pixtral-large-2411'
    ),
  }),

  # Hugging Face models suggested in:
  # https://huggingface.co/docs/inference-providers/en/tasks/chat-completion
  'huggingface': MappingProxyType({
    'gemma-2-2b-it': types.ProviderModelType(
      provider='huggingface',
      model='gemma-2-2b-it',
      provider_model_identifier='google/gemma-2-2b-it'
    ),
    'meta-llama-3.1-8b-it': types.ProviderModelType(
      provider='huggingface',
      model='meta-llama-3.1-8b-it',
      provider_model_identifier='meta-llama/Meta-Llama-3.1-8B-Instruct'
    ),
    'phi-4': types.ProviderModelType(
      provider='huggingface',
      model='phi-4',
      provider_model_identifier='microsoft/phi-4'
    ),
    'qwen3-32b': types.ProviderModelType(
      provider='huggingface',
      model='qwen3-32b',
      provider_model_identifier='Qwen/Qwen3-32B'
    ),
    'deepseek-r1': types.ProviderModelType(
      provider='huggingface',
      model='deepseek-r1',
      provider_model_identifier='deepseek-ai/DeepSeek-R1'
    ),
    'deepseek-v3': types.ProviderModelType(
      provider='huggingface',
      model='deepseek-v3',
      provider_model_identifier='deepseek-ai/DeepSeek-V3'
    ),
  }),

  'deepseek': MappingProxyType({
    'deepseek-v3': types.ProviderModelType(
      provider='deepseek',
      model='deepseek-v3',
      provider_model_identifier='deepseek-chat'
    ),
    'deepseek-r1': types.ProviderModelType(
      provider='deepseek',
      model='deepseek-r1',
      provider_model_identifier='deepseek-reasoner'
    ),
  }),

  'grok': MappingProxyType({
    'grok-3-beta': types.ProviderModelType(
      provider='grok',
      model='grok-3-beta',
      provider_model_identifier='grok-3-beta'
    ),
    'grok-3-fast-beta': types.ProviderModelType(
      provider='grok',
      model='grok-3-fast-beta',
      provider_model_identifier='grok-3-fast-beta'
    ),
    'grok-3-mini-beta': types.ProviderModelType(
      provider='grok',
      model='grok-3-mini-beta',
      provider_model_identifier='grok-3-mini-beta'
    ),
    'grok-3-mini-fast-beta': types.ProviderModelType(
      provider='grok',
      model='grok-3-mini-fast-beta',
      provider_model_identifier='grok-3-mini-fast-beta'
    ),
  }),
})


GENERATE_TEXT_MODELS: Dict[
    str, Dict[str, types.ProviderModelType]] = MappingProxyType({
  'mock_provider': MappingProxyType({
    'mock_model': ALL_MODELS['mock_provider']['mock_model'],
  }),

  'mock_failing_provider': MappingProxyType({
    'mock_failing_model': ALL_MODELS[
        'mock_failing_provider']['mock_failing_model'],
  }),

  'mock_slow_provider': MappingProxyType({
    'mock_slow_model': ALL_MODELS['mock_slow_provider']['mock_slow_model'],
  }),

  'openai': MappingProxyType({
    'gpt-4.1': ALL_MODELS['openai']['gpt-4.1'],
    'gpt-4.1-mini': ALL_MODELS['openai']['gpt-4.1-mini'],
    'gpt-4.1-nano': ALL_MODELS['openai']['gpt-4.1-nano'],
    'gpt-4.5-preview': ALL_MODELS['openai']['gpt-4.5-preview'],
    'gpt-4o': ALL_MODELS['openai']['gpt-4o'],
    'gpt-4o-mini': ALL_MODELS['openai']['gpt-4o-mini'],
    'o1': ALL_MODELS['openai']['o1'],
    'o1-pro': ALL_MODELS['openai']['o1-pro'],
    'o3': ALL_MODELS['openai']['o3'],
    'o4-mini': ALL_MODELS['openai']['o4-mini'],
    'o3-mini': ALL_MODELS['openai']['o3-mini'],
    'o1-mini': ALL_MODELS['openai']['o1-mini'],
    'gpt-4o-mini-search-preview': ALL_MODELS['openai']['gpt-4o-mini-search-preview'],
    'gpt-4o-search-preview': ALL_MODELS['openai']['gpt-4o-search-preview'],
    'chatgpt-4o-latest': ALL_MODELS['openai']['chatgpt-4o-latest'],
    'gpt-4-turbo': ALL_MODELS['openai']['gpt-4-turbo'],
    'gpt-4': ALL_MODELS['openai']['gpt-4'],
    'gpt-3.5-turbo': ALL_MODELS['openai']['gpt-3.5-turbo'],
  }),

  'claude': MappingProxyType({
    'sonnet': ALL_MODELS['claude']['sonnet'],
    'haiku': ALL_MODELS['claude']['haiku'],
    '3.5-sonnet-v2': ALL_MODELS['claude']['3.5-sonnet-v2'],
    '3.5-sonnet': ALL_MODELS['claude']['3.5-sonnet'],
    'opus': ALL_MODELS['claude']['opus'],
    '3-sonnet': ALL_MODELS['claude']['3-sonnet'],
    '3-haiku': ALL_MODELS['claude']['3-haiku'],
  }),

  'gemini': MappingProxyType({
    'gemini-1.5-pro': ALL_MODELS['gemini']['gemini-1.5-pro'],
    'gemini-1.5-flash-8b': ALL_MODELS['gemini']['gemini-1.5-flash-8b'],
    'gemini-1.5-flash': ALL_MODELS['gemini']['gemini-1.5-flash'],
    'gemini-2.0-flash-lite': ALL_MODELS['gemini']['gemini-2.0-flash-lite'],
    'gemini-2.0-flash': ALL_MODELS['gemini']['gemini-2.0-flash'],
    'gemini-2.5-flash-preview': ALL_MODELS['gemini']['gemini-2.5-flash-preview'],
    'gemini-2.5-pro-preview': ALL_MODELS['gemini']['gemini-2.5-pro-preview'],
  }),

  'cohere': MappingProxyType({
    'command-a': ALL_MODELS['cohere']['command-a'],
    'command-r7b': ALL_MODELS['cohere']['command-r7b'],
    'command-r-plus': ALL_MODELS['cohere']['command-r-plus'],
    'command-r-08-2024': ALL_MODELS['cohere']['command-r-08-2024'],
    'command-r': ALL_MODELS['cohere']['command-r'],
    'command': ALL_MODELS['cohere']['command'],
    'command-nightly': ALL_MODELS['cohere']['command-nightly'],
    'command-light': ALL_MODELS['cohere']['command-light'],
    'command-light-nightly': ALL_MODELS['cohere']['command-light-nightly'],
  }),

  'databricks': MappingProxyType({
    'llama-4-maverick': ALL_MODELS['databricks']['llama-4-maverick'],
    'claude-3-7-sonnet': ALL_MODELS['databricks']['claude-3-7-sonnet'],
    'meta-llama-3-1-8b-it': ALL_MODELS['databricks']['meta-llama-3-1-8b-it'],
    'meta-llama-3-3-70b-it': ALL_MODELS['databricks']['meta-llama-3-3-70b-it'],
    # TODO: This is extremely slow model. Until better filtering, it is not
    # included in the list.
    # 'meta-llama-3-1-405b-it': ALL_MODELS['databricks']['meta-llama-3-1-405b-it'],
    'dbrx-it': ALL_MODELS['databricks']['dbrx-it'],
    'mixtral-8x7b-it': ALL_MODELS['databricks']['mixtral-8x7b-it'],
  }),

  'mistral': MappingProxyType({
    'codestral': ALL_MODELS['mistral']['codestral'],
    'ministral-3b': ALL_MODELS['mistral']['ministral-3b'],
    'ministral-8b': ALL_MODELS['mistral']['ministral-8b'],
    'mistral-large': ALL_MODELS['mistral']['mistral-large'],
    'mistral-medium': ALL_MODELS['mistral']['mistral-medium'],
    'mistral-saba': ALL_MODELS['mistral']['mistral-saba'],
    'mistral-small': ALL_MODELS['mistral']['mistral-small'],
    'open-mistral-7b': ALL_MODELS['mistral']['open-mistral-7b'],
    'open-mistral-nemo': ALL_MODELS['mistral']['open-mistral-nemo'],
    'open-mixtral-8x22b': ALL_MODELS['mistral']['open-mixtral-8x22b'],
    'open-mixtral-8x7b': ALL_MODELS['mistral']['open-mixtral-8x7b'],
    'pixtral-12b': ALL_MODELS['mistral']['pixtral-12b'],
    'pixtral-large': ALL_MODELS['mistral']['pixtral-large'],
  }),

  'huggingface': MappingProxyType({
    'gemma-2-2b-it': ALL_MODELS['huggingface']['gemma-2-2b-it'],
    'meta-llama-3.1-8b-it': ALL_MODELS['huggingface']['meta-llama-3.1-8b-it'],
    'phi-4': ALL_MODELS['huggingface']['phi-4'],
    'qwen3-32b': ALL_MODELS['huggingface']['qwen3-32b'],
    'deepseek-r1': ALL_MODELS['huggingface']['deepseek-r1'],
    'deepseek-v3': ALL_MODELS['huggingface']['deepseek-v3'],
  }),

  'deepseek': MappingProxyType({
    'deepseek-v3': ALL_MODELS['deepseek']['deepseek-v3'],
    'deepseek-r1': ALL_MODELS['deepseek']['deepseek-r1'],
  }),

  'grok': MappingProxyType({
    'grok-3-beta': ALL_MODELS['grok']['grok-3-beta'],
    'grok-3-fast-beta': ALL_MODELS['grok']['grok-3-fast-beta'],
    'grok-3-mini-beta': ALL_MODELS['grok']['grok-3-mini-beta'],
    'grok-3-mini-fast-beta': ALL_MODELS['grok']['grok-3-mini-fast-beta'],
  }),
})

SMALL_GENERATE_TEXT_MODELS: Dict[
    str, Dict[str, types.ProviderModelType]] = MappingProxyType({
  'claude': MappingProxyType({
    '3-haiku': ALL_MODELS['claude']['3-haiku'],
  }),
  'cohere': MappingProxyType({
    'command-light': ALL_MODELS['cohere']['command-light'],
    'command-light-nightly': ALL_MODELS['cohere']['command-light-nightly'],
    'command-r': ALL_MODELS['cohere']['command-r'],
    'command-r-08-2024': ALL_MODELS['cohere']['command-r-08-2024'],
    'command-r7b': ALL_MODELS['cohere']['command-r7b'],
  }),
  'deepseek': MappingProxyType({
    'deepseek-v3': ALL_MODELS['deepseek']['deepseek-v3'],
  }),
  'gemini': MappingProxyType({
    'gemini-1.5-flash': ALL_MODELS['gemini']['gemini-1.5-flash'],
    'gemini-1.5-flash-8b': ALL_MODELS['gemini']['gemini-1.5-flash-8b'],
    'gemini-2.0-flash': ALL_MODELS['gemini']['gemini-2.0-flash'],
    'gemini-2.0-flash-lite': ALL_MODELS['gemini']['gemini-2.0-flash-lite'],
  }),
  'grok': MappingProxyType({
    'grok-3-mini-beta': ALL_MODELS['grok']['grok-3-mini-beta'],
  }),
  'mistral': MappingProxyType({
    'codestral': ALL_MODELS['mistral']['codestral'],
    'ministral-3b': ALL_MODELS['mistral']['ministral-3b'],
    'ministral-8b': ALL_MODELS['mistral']['ministral-8b'],
    'mistral-saba': ALL_MODELS['mistral']['mistral-saba'],
    'mistral-small': ALL_MODELS['mistral']['mistral-small'],
    'open-mistral-7b': ALL_MODELS['mistral']['open-mistral-7b'],
    'open-mistral-nemo': ALL_MODELS['mistral']['open-mistral-nemo'],
    'open-mixtral-8x7b': ALL_MODELS['mistral']['open-mixtral-8x7b'],
    'pixtral-12b': ALL_MODELS['mistral']['pixtral-12b'],
  }),
  'openai': MappingProxyType({
    'gpt-4.1-nano': ALL_MODELS['openai']['gpt-4.1-nano'],
    'gpt-4o-mini': ALL_MODELS['openai']['gpt-4o-mini'],
    'gpt-4o-mini-search-preview': ALL_MODELS['openai']['gpt-4o-mini-search-preview'],
  }),
})

MEDIUM_GENERATE_TEXT_MODELS: Dict[
    str, Dict[str, types.ProviderModelType]] = MappingProxyType({
  'claude': MappingProxyType({
    'haiku': ALL_MODELS['claude']['haiku'],
  }),
  'cohere': MappingProxyType({
    'command': ALL_MODELS['cohere']['command'],
    'command-nightly': ALL_MODELS['cohere']['command-nightly'],
  }),
  'deepseek': MappingProxyType({
    'deepseek-r1': ALL_MODELS['deepseek']['deepseek-r1'],
  }),
  'gemini': MappingProxyType({
    'gemini-1.5-pro': ALL_MODELS['gemini']['gemini-1.5-pro'],
    'gemini-2.5-flash-preview': ALL_MODELS['gemini']['gemini-2.5-flash-preview'],
  }),
  'grok': MappingProxyType({
    'grok-3-mini-fast-beta': ALL_MODELS['grok']['grok-3-mini-fast-beta'],
  }),
  'huggingface': MappingProxyType({
    'gemma-2-2b-it': ALL_MODELS['huggingface']['gemma-2-2b-it'],
    'meta-llama-3.1-8b-it': ALL_MODELS['huggingface']['meta-llama-3.1-8b-it'],
  }),
  'mistral': MappingProxyType({
    'mistral-large': ALL_MODELS['mistral']['mistral-large'],
    'open-mixtral-8x22b': ALL_MODELS['mistral']['open-mixtral-8x22b'],
    'pixtral-large': ALL_MODELS['mistral']['pixtral-large'],
  }),
  'openai': MappingProxyType({
    'gpt-3.5-turbo': ALL_MODELS['openai']['gpt-3.5-turbo'],
    'gpt-4.1-mini': ALL_MODELS['openai']['gpt-4.1-mini'],
    'o1-mini': ALL_MODELS['openai']['o1-mini'],
    'o3-mini': ALL_MODELS['openai']['o3-mini'],
    'o4-mini': ALL_MODELS['openai']['o4-mini'],
  }),
})

LARGE_GENERATE_TEXT_MODELS: Dict[
    str, Dict[str, types.ProviderModelType]] = MappingProxyType({
  'claude': MappingProxyType({
    'opus': ALL_MODELS['claude']['opus'],
    'sonnet': ALL_MODELS['claude']['sonnet'],
  }),
  'cohere': MappingProxyType({
    'command-a': ALL_MODELS['cohere']['command-a'],
    'command-r-plus': ALL_MODELS['cohere']['command-r-plus'],
  }),
  'databricks': MappingProxyType({
    'claude-3-7-sonnet': ALL_MODELS['databricks']['claude-3-7-sonnet'],
    'dbrx-it': ALL_MODELS['databricks']['dbrx-it'],
    'llama-4-maverick': ALL_MODELS['databricks']['llama-4-maverick'],
    'meta-llama-3-1-8b-it': ALL_MODELS['databricks']['meta-llama-3-1-8b-it'],
    'meta-llama-3-3-70b-it': ALL_MODELS['databricks']['meta-llama-3-3-70b-it'],
    'mixtral-8x7b-it': ALL_MODELS['databricks']['mixtral-8x7b-it'],
  }),
  'gemini': MappingProxyType({
    'gemini-2.5-pro-preview': ALL_MODELS['gemini']['gemini-2.5-pro-preview'],
  }),
  'grok': MappingProxyType({
    'grok-3-beta': ALL_MODELS['grok']['grok-3-beta'],
    'grok-3-fast-beta': ALL_MODELS['grok']['grok-3-fast-beta'],
  }),
  'huggingface': MappingProxyType({
    'deepseek-r1': ALL_MODELS['huggingface']['deepseek-r1'],
    'deepseek-v3': ALL_MODELS['huggingface']['deepseek-v3'],
    'phi-4': ALL_MODELS['huggingface']['phi-4'],
    'qwen3-32b': ALL_MODELS['huggingface']['qwen3-32b'],
  }),
  'openai': MappingProxyType({
    'chatgpt-4o-latest': ALL_MODELS['openai']['chatgpt-4o-latest'],
    'gpt-4': ALL_MODELS['openai']['gpt-4'],
    'gpt-4-turbo': ALL_MODELS['openai']['gpt-4-turbo'],
    'gpt-4.1': ALL_MODELS['openai']['gpt-4.1'],
    'gpt-4.5-preview': ALL_MODELS['openai']['gpt-4.5-preview'],
    'gpt-4o': ALL_MODELS['openai']['gpt-4o'],
    'gpt-4o-search-preview': ALL_MODELS['openai']['gpt-4o-search-preview'],
    'o1': ALL_MODELS['openai']['o1'],
    'o1-pro': ALL_MODELS['openai']['o1-pro'],
    'o3': ALL_MODELS['openai']['o3'],
  }),
})

LARGEST_GENERATE_TEXT_MODELS: Dict[
    str, Dict[str, types.ProviderModelType]] = MappingProxyType({
  'claude': MappingProxyType({
    'sonnet': ALL_MODELS['claude']['sonnet'],
  }),
  'cohere': MappingProxyType({
    'command-a': ALL_MODELS['cohere']['command-a'],
  }),
  'databricks': MappingProxyType({
    'dbrx-it': ALL_MODELS['databricks']['dbrx-it'],
    'meta-llama-3-3-70b-it': ALL_MODELS['databricks']['meta-llama-3-3-70b-it'],
  }),
  'deepseek': MappingProxyType({
    'deepseek-r1': ALL_MODELS['deepseek']['deepseek-r1'],
  }),
  'gemini': MappingProxyType({
    'gemini-2.5-pro-preview': ALL_MODELS['gemini']['gemini-2.5-pro-preview'],
  }),
  'grok': MappingProxyType({
    'grok-3-beta': ALL_MODELS['grok']['grok-3-beta'],
  }),
  'huggingface': MappingProxyType({
    'phi-4': ALL_MODELS['huggingface']['phi-4'],
    'qwen3-32b': ALL_MODELS['huggingface']['qwen3-32b'],
  }),
  'mistral': MappingProxyType({
    'mistral-large': ALL_MODELS['mistral']['mistral-large'],
  }),
  'openai': MappingProxyType({
    'gpt-4.1': ALL_MODELS['openai']['gpt-4.1'],
  }),
})

PREFERRED_DEFAULT_MODELS: List[types.ProviderModelType] = [
  ALL_MODELS['openai']['gpt-4o-mini'],
  ALL_MODELS['gemini']['gemini-2.0-flash'],
  ALL_MODELS['claude']['haiku'],
  ALL_MODELS['grok']['grok-3-mini-fast-beta'],
  ALL_MODELS['cohere']['command-r'],
  ALL_MODELS['mistral']['mistral-small'],
  ALL_MODELS['deepseek']['deepseek-v3'],
  ALL_MODELS['huggingface']['gemma-2-2b-it'],
  ALL_MODELS['databricks']['llama-4-maverick'],
]


def get_provider_model_config(
    model_identifier: types.ProviderModelIdentifierType
) -> types.ProviderModelType:
  if type_utils.is_provider_model_tuple(model_identifier):
    return ALL_MODELS[model_identifier[0]][model_identifier[1]]
  else:
    return model_identifier



import copy
import datetime
import traceback
import functools
import math
from typing import Any, Callable, Dict, List, Optional
import proxai.types as types
import proxai.logging.utils as logging_utils
import proxai.caching.query_cache as query_cache
import proxai.type_utils as type_utils
import proxai.stat_types as stats_type
import proxai.serializers.hash_serializer as hash_serializer
import proxai.connections.proxdash as proxdash
import proxai.connectors.model_configs as model_configs
import proxai.connectors.model_pricing_configs as model_pricing_configs
import proxai.connectors.model_feature_configs as model_feature_configs
import proxai.state_controllers.state_controller as state_controller

_PROVIDER_MODEL_STATE_PROPERTY = '_provider_model_state'


class ProviderModelConnector(state_controller.StateControlled):
  _provider_model: Optional[types.ProviderModelType]
  _run_type: Optional[types.RunType]
  _get_run_type: Optional[Callable[[], types.RunType]]
  _strict_feature_test: Optional[bool]
  _get_strict_feature_test: Optional[Callable[[], bool]]
  _query_cache_manager: Optional[query_cache.QueryCacheManager]
  _get_query_cache_manager: Optional[
      Callable[[], query_cache.QueryCacheManager]]
  _api: Optional[Any]
  _stats: Optional[Dict[str, stats_type.RunStats]]
  _logging_options: Optional[types.LoggingOptions]
  _get_logging_options: Optional[Dict]
  _proxdash_connection: Optional[proxdash.ProxDashConnection]
  _get_proxdash_connection: Optional[
      Callable[[bool], proxdash.ProxDashConnection]]
  _provider_model_state: Optional[types.ProviderModelState]

  def __init__(
      self,
      provider_model: Optional[types.ProviderModelType] = None,
      run_type: Optional[types.RunType] = None,
      get_run_type: Optional[Callable[[], types.RunType]] = None,
      strict_feature_test: Optional[bool] = None,
      get_strict_feature_test: Optional[Callable[[], bool]] = None,
      query_cache_manager: Optional[query_cache.QueryCacheManager] = None,
      get_query_cache_manager: Optional[
          Callable[[], query_cache.QueryCacheManager]] = None,
      logging_options: Optional[types.LoggingOptions] = None,
      get_logging_options: Optional[Callable[[], types.LoggingOptions]] = None,
      proxdash_connection: Optional[proxdash.ProxDashConnection] = None,
      get_proxdash_connection: Optional[
          Callable[[bool], proxdash.ProxDashConnection]] = None,
      init_state: Optional[types.ProviderModelState] = None,
      stats: Optional[Dict[str, stats_type.RunStats]] = None):

    if init_state and (
        provider_model is not None or
        run_type is not None or
        get_run_type is not None or
        strict_feature_test is not None or
        get_strict_feature_test is not None or
        query_cache_manager is not None or
        get_query_cache_manager is not None or
        logging_options is not None or
        get_logging_options is not None or
        proxdash_connection is not None or
        get_proxdash_connection is not None or
        stats is not None):
      raise ValueError(
          'init_state and other parameters cannot be set at the same time.')

    super().__init__(
        provider_model=provider_model,
        run_type=run_type,
        get_run_type=get_run_type,
        strict_feature_test=strict_feature_test,
        get_strict_feature_test=get_strict_feature_test,
        query_cache_manager=query_cache_manager,
        get_query_cache_manager=get_query_cache_manager,
        logging_options=logging_options,
        get_logging_options=get_logging_options,
        proxdash_connection=proxdash_connection,
        get_proxdash_connection=get_proxdash_connection)

    self.init_state()

    if init_state:
      if init_state.provider_model is None:
        raise ValueError('provider_model needs to be set in init_state.')
      if init_state.provider_model.provider != self.get_provider_name():
        raise ValueError(
            'provider_model needs to be same with the class provider name.\n'
            f'provider_model: {init_state.provider_model}\n'
            f'class provider name: {self.get_provider_name()}')
      self.load_state(init_state)
    else:
      initial_state = self.get_state()

      self._get_run_type = get_run_type
      self._get_strict_feature_test = get_strict_feature_test
      self._get_query_cache_manager = get_query_cache_manager
      self._get_logging_options = get_logging_options
      self._get_proxdash_connection = get_proxdash_connection

      self.provider_model = provider_model
      self.run_type = run_type
      self.strict_feature_test = strict_feature_test
      self.query_cache_manager = query_cache_manager
      self.logging_options = logging_options
      self.proxdash_connection = proxdash_connection
      self._stats = stats

      self.handle_changes(initial_state, self.get_state())

  def get_internal_state_property_name(self):
    return _PROVIDER_MODEL_STATE_PROPERTY

  def get_internal_state_type(self):
    return types.ProviderModelState

  def handle_changes(
      self,
      old_state: types.ProviderModelState,
      current_state: types.ProviderModelState):
    result_state = copy.deepcopy(old_state)
    if current_state.provider_model is not None:
      result_state.provider_model = current_state.provider_model
    if current_state.run_type is not None:
      result_state.run_type = current_state.run_type
    if current_state.strict_feature_test is not None:
      result_state.strict_feature_test = current_state.strict_feature_test
    if current_state.logging_options is not None:
      result_state.logging_options = current_state.logging_options
    if current_state.proxdash_connection is not None:
      result_state.proxdash_connection = (
          current_state.proxdash_connection)

    if result_state.provider_model is None:
      raise ValueError(
          'Provider model is not set for both old and new states. '
          'This creates an invalid state change.')

    if result_state.provider_model.provider != self.get_provider_name():
      raise ValueError(
          'Provider needs to be same with the class provider name.\n'
          f'provider_model: {result_state.provider_model}\n'
          f'class provider name: {self.get_provider_name()}')

    if result_state.logging_options is None:
      raise ValueError(
          'Logging options are not set for both old and new states. '
          'This creates an invalid state change.')
    if result_state.proxdash_connection is None:
      raise ValueError(
          'ProxDash connection is not set for both old and new states. '
          'This creates an invalid state change.')

  @property
  def api(self):
    if not getattr(self, '_api', None):
      if self.run_type == types.RunType.PRODUCTION:
        self._api = self.init_model()
      else:
        self._api = self.init_mock_model()
    return self._api

  @api.setter
  def api(self, value):
    raise ValueError('api should not be set directly.')

  @property
  def provider_model(self):
    return self.get_property_value('provider_model')

  @provider_model.setter
  def provider_model(self, value):
    self.set_property_value('provider_model', value)

  @property
  def run_type(self):
    return self.get_property_value('run_type')

  @run_type.setter
  def run_type(self, value):
    self.set_property_value('run_type', value)

  @property
  def strict_feature_test(self):
    return self.get_property_value('strict_feature_test')

  @strict_feature_test.setter
  def strict_feature_test(self, value):
    self.set_property_value('strict_feature_test', value)

  @property
  def query_cache_manager(self):
    return self.get_state_controlled_property_value('query_cache_manager')

  @query_cache_manager.setter
  def query_cache_manager(self, value):
    self.set_state_controlled_property_value('query_cache_manager', value)

  def query_cache_manager_deserializer(
      self,
      state_value: types.QueryCacheManagerState
  ) -> query_cache.QueryCacheManager:
    return query_cache.QueryCacheManager(init_state=state_value)

  @property
  def logging_options(self):
    return self.get_property_value('logging_options')

  @logging_options.setter
  def logging_options(self, value):
    self.set_property_value('logging_options', value)

  @property
  def proxdash_connection(self):
    return self.get_state_controlled_property_value('proxdash_connection')

  @proxdash_connection.setter
  def proxdash_connection(self, value):
    self.set_state_controlled_property_value('proxdash_connection', value)

  def proxdash_connection_deserializer(
      self,
      state_value: types.ProxDashConnectionState
  ) -> proxdash.ProxDashConnection:
    return proxdash.ProxDashConnection(init_state=state_value)

  def feature_fail(
      self,
      message: str,
      query_record: Optional[types.QueryRecord] = None):
    if self.strict_feature_test:
      logging_utils.log_message(
          type=types.LoggingType.ERROR,
          logging_options=self.logging_options,
          query_record=query_record,
          message=message)
      raise Exception(message)
    else:
      logging_utils.log_message(
          type=types.LoggingType.WARNING,
          logging_options=self.logging_options,
          query_record=query_record,
          message=message)

  def feature_check(self, query_record: types.QueryRecord) -> types.QueryRecord:
    query_record = copy.deepcopy(query_record)
    if (self.provider_model.provider
        not in model_feature_configs.GENERATE_TEXT_FEATURES):
      return query_record

    model_features = model_feature_configs.GENERATE_TEXT_FEATURES[
        self.provider_model.provider]
    if self.provider_model.model not in model_features:
      return query_record

    model_feature_config = model_features[self.provider_model.model]
    for feature in model_feature_config.not_supported_features:
      if getattr(query_record, feature) is not None:
        self.feature_fail(
            message=f'{self.provider_model.model} does not support {feature}.',
            query_record=query_record)
        setattr(query_record, feature, None)
    return query_record

  def get_token_count_estimate(
      self,
      prompt: Optional[str] = None,
      messages: Optional[List[Dict[str, str]]] = None) -> int:
    total = 0
    def _get_token_count_estimate_from_prompt(prompt: str) -> int:
      return math.ceil(max(
          len(prompt) / 4,
          len(prompt.strip().split()) * 1.3))
    if prompt is not None:
      total += _get_token_count_estimate_from_prompt(prompt)
    if messages is not None:
      total += 2
      for message in messages:
        total += _get_token_count_estimate_from_prompt(
            message.get("content", "")) + 4
    return total

  def get_estimated_cost(self, logging_record: types.LoggingRecord):
    query_token_count = logging_record.query_record.token_count
    if type(query_token_count) != int:
      query_token_count = 0
    response_token_count = logging_record.response_record.token_count
    if type(response_token_count) != int:
      response_token_count = 0
    return model_pricing_configs.get_provider_model_cost(
        provider_model_identifier=logging_record.query_record.provider_model,
        query_token_count=query_token_count,
        response_token_count=response_token_count)

  def _update_stats(self, logging_record: types.LoggingRecord):
    if getattr(self, '_stats', None) is None:
      return
    provider_stats = stats_type.BaseProviderStats()
    cache_stats = stats_type.BaseCacheStats()
    query_token_count = logging_record.query_record.token_count
    if type(query_token_count) != int:
      query_token_count = 0
    response_token_count = logging_record.response_record.token_count
    if type(response_token_count) != int:
      response_token_count = 0
    if logging_record.response_source == types.ResponseSource.PROVIDER:
      provider_stats.total_queries = 1
      if logging_record.response_record.response:
        provider_stats.total_successes = 1
      else:
        provider_stats.total_fails = 1
      provider_stats.total_token_count = (
          query_token_count + response_token_count)
      provider_stats.total_query_token_count = query_token_count
      provider_stats.total_response_token_count = response_token_count
      provider_stats.total_response_time = (
          logging_record.response_record.response_time.total_seconds())
      provider_stats.estimated_cost = (
          logging_record.response_record.estimated_cost)
      provider_stats.total_cache_look_fail_reasons = {
          logging_record.look_fail_reason: 1}
    elif logging_record.response_source == types.ResponseSource.CACHE:
      cache_stats.total_cache_hit = 1
      if logging_record.response_record.response:
        cache_stats.total_success_return = 1
      else:
        cache_stats.total_fail_return = 1
      cache_stats.saved_token_count = (
          query_token_count + response_token_count)
      cache_stats.saved_query_token_count = query_token_count
      cache_stats.saved_response_token_count = response_token_count
      cache_stats.saved_total_response_time = (
          logging_record.response_record.response_time.total_seconds())
      cache_stats.saved_estimated_cost = (
          logging_record.response_record.estimated_cost)
    else:
      raise ValueError(
        f'Invalid response source.\n{logging_record.response_source}')

    provider_model_stats = stats_type.ProviderModelStats(
        provider_model=self.provider_model,
        provider_stats=provider_stats,
        cache_stats=cache_stats)
    self._stats[stats_type.GlobalStatType.RUN_TIME] += provider_model_stats
    self._stats[stats_type.GlobalStatType.SINCE_CONNECT] += provider_model_stats

  def _update_proxdash(self, logging_record: types.LoggingRecord):
    if not self.proxdash_connection:
      return
    try:
      self.proxdash_connection.upload_logging_record(logging_record)
    except Exception as e:
      logging_utils.log_proxdash_message(
          logging_options=self.logging_options,
          message=(
              'ProxDash upload_logging_record failed.\n'
              f'Error message: {e}\n'
              f'Traceback: {traceback.format_exc()}'),
          type=types.LoggingType.ERROR)

  def generate_text(
      self,
      prompt: Optional[str] = None,
      system: Optional[str] = None,
      messages: Optional[types.MessagesType] = None,
      max_tokens: Optional[int] = None,
      temperature: Optional[float] = None,
      stop: Optional[types.StopType] = None,
      provider_model: Optional[types.ProviderModelIdentifierType] = None,
      use_cache: bool = True,
      unique_response_limit: Optional[int] = None) -> types.LoggingRecord:
    if prompt != None and messages != None:
      raise ValueError('prompt and messages cannot be set at the same time.')
    if messages != None:
      type_utils.check_messages_type(messages)

    if provider_model is not None:
      provider_model = model_configs.get_provider_model_config(
          model_identifier=provider_model)
      if provider_model != self.provider_model:
        raise ValueError(
            'provider_model does not match the connector provider_model.'
            f'provider_model: {provider_model}\n'
            f'connector provider_model: {self.provider_model}')

    start_utc_date = datetime.datetime.now(datetime.timezone.utc)
    query_record = types.QueryRecord(
        call_type=types.CallType.GENERATE_TEXT,
        provider_model=self.provider_model,
        prompt=prompt,
        system=system,
        messages=messages,
        max_tokens=max_tokens,
        temperature=temperature,
        stop=stop,
        token_count=self.get_token_count_estimate(
            prompt=prompt,
            messages=messages))

    updated_query_record = self.feature_check(query_record=query_record)

    look_fail_reason = None
    if self.query_cache_manager and use_cache:
      cache_look_result = None
      response_record = None
      try:
        cache_look_result = self.query_cache_manager.look(
            updated_query_record,
            unique_response_limit=unique_response_limit)
        if cache_look_result.query_response:
          response_record = cache_look_result.query_response
      except Exception as e:
        pass
      if response_record:
        response_record.end_utc_date = datetime.datetime.now(
            datetime.timezone.utc)
        response_record.start_utc_date = (
            response_record.end_utc_date - response_record.response_time)
        response_record.local_time_offset_minute = (
            datetime.datetime.now().astimezone().utcoffset().total_seconds()
            // 60) * -1
        logging_record = types.LoggingRecord(
            query_record=query_record,
            response_record=response_record,
            response_source=types.ResponseSource.CACHE)
        logging_record.response_record.estimated_cost = (
            self.get_estimated_cost(logging_record=logging_record))
        logging_utils.log_logging_record(
            logging_options=self.logging_options,
            logging_record=logging_record)
        self._update_stats(logging_record=logging_record)
        self._update_proxdash(logging_record=logging_record)
        return logging_record
      look_fail_reason = cache_look_result.look_fail_reason
      logging_record = types.LoggingRecord(
          query_record=query_record,
          look_fail_reason=look_fail_reason,
          response_source=types.ResponseSource.CACHE)
      logging_utils.log_logging_record(
          logging_options=self.logging_options,
          logging_record=logging_record)

    response, error, error_traceback = None, None, None
    try:
      response = self.generate_text_proc(query_record=updated_query_record)
    except Exception as e:
      error_traceback = traceback.format_exc()
      error = e

    if response != None:
      query_response_record = functools.partial(
          types.QueryResponseRecord,
          response=response,
          token_count=self.get_token_count_estimate(prompt=response))
    else:
      query_response_record = functools.partial(
          types.QueryResponseRecord,
          error=str(error),
          error_traceback=error_traceback)
    response_record = query_response_record(
        start_utc_date=start_utc_date,
        end_utc_date=datetime.datetime.now(datetime.timezone.utc),
        local_time_offset_minute=(
            datetime.datetime.now().astimezone().utcoffset().total_seconds()
            // 60) * -1,
        response_time=(
            datetime.datetime.now(datetime.timezone.utc) - start_utc_date))

    if self.query_cache_manager and use_cache:
      self.query_cache_manager.cache(
          query_record=updated_query_record,
          response_record=response_record,
          unique_response_limit=unique_response_limit)

    logging_record = types.LoggingRecord(
        query_record=query_record,
        response_record=response_record,
        look_fail_reason=look_fail_reason,
        response_source=types.ResponseSource.PROVIDER)
    logging_record.response_record.estimated_cost = (
        self.get_estimated_cost(logging_record=logging_record))
    logging_utils.log_logging_record(
        logging_options=self.logging_options,
        logging_record=logging_record)
    self._update_stats(logging_record=logging_record)
    self._update_proxdash(logging_record=logging_record)
    return logging_record

  def get_provider_name(self):
    raise NotImplementedError

  def init_model(self):
    raise NotImplementedError

  def init_mock_model(self):
    raise NotImplementedError

  def generate_text_proc(self, query_record: types.QueryRecord) -> dict:
    raise NotImplementedError



import proxai.types as types


def get_query_record_hash(query_record: types.QueryRecord) -> str:
  POWER = 65535
  PRIME = 110589769251386569
  signature_str = ''
  if query_record.call_type != None:
    signature_str += query_record.call_type + chr(255)
  if query_record.provider_model != None:
    signature_str += query_record.provider_model.provider + chr(255)
    signature_str += query_record.provider_model.model + chr(255)
    signature_str += (
        query_record.provider_model.provider_model_identifier + chr(255))
  if query_record.prompt != None:
    signature_str += query_record.prompt + chr(255)
  if query_record.system != None:
    signature_str += query_record.system + chr(255)
  if query_record.messages != None:
    for message in query_record.messages:
      signature_str += 'role:'+message['role'] + chr(255)
      signature_str += 'content:'+message['content'] + chr(255)
  if query_record.max_tokens != None:
    signature_str += str(query_record.max_tokens) + chr(255)
  if query_record.temperature != None:
    signature_str += str(query_record.temperature) + chr(255)
  if query_record.stop != None:
    signature_str += str(query_record.stop) + chr(255)
  hash_val = 0
  for char in signature_str:
    hash_val = (hash_val * POWER + ord(char)) % PRIME
  return str(hash_val)



import math
from types import MappingProxyType
from typing import Dict
import proxai.types as types
import proxai.connectors.model_configs as model_configs

GENERATE_TEXT_PRICING: Dict[
    str, Dict[str, types.ProviderModelPricingType]] = MappingProxyType({
  # Mock provider
  'mock_provider': MappingProxyType({
    'mock_model': types.ProviderModelPricingType(
      per_response_token_cost=1.0,
      per_query_token_cost=2.0,
    ),
  }),

  # Mock failing provider
  'mock_failing_provider': MappingProxyType({
    'mock_failing_model': types.ProviderModelPricingType(
      per_response_token_cost=3.0,
      per_query_token_cost=4.0,
    ),
  }),

  # OpenAI
  'openai': MappingProxyType({
    'gpt-4.1': types.ProviderModelPricingType(
      per_response_token_cost=2.00,
      per_query_token_cost=8.00,
    ),
    'gpt-4.1-mini': types.ProviderModelPricingType(
      per_response_token_cost=0.40,
      per_query_token_cost=1.60,
    ),
    'gpt-4.1-nano': types.ProviderModelPricingType(
      per_response_token_cost=0.10,
      per_query_token_cost=0.40,
    ),
    'gpt-4.5-preview': types.ProviderModelPricingType(
      per_response_token_cost=75.00,
      per_query_token_cost=150.00,
    ),
    'gpt-4o': types.ProviderModelPricingType(
      per_response_token_cost=2.50,
      per_query_token_cost=10.00,
    ),
    'gpt-4o-audio-preview': types.ProviderModelPricingType(
      per_response_token_cost=2.50,
      per_query_token_cost=10.00,
    ),
    'gpt-4o-realtime-preview': types.ProviderModelPricingType(
      per_response_token_cost=5.00,
      per_query_token_cost=20.00,
    ),
    'gpt-4o-mini': types.ProviderModelPricingType(
      per_response_token_cost=0.15,
      per_query_token_cost=0.60,
    ),
    'gpt-4o-mini-audio-preview': types.ProviderModelPricingType(
      per_response_token_cost=0.15,
      per_query_token_cost=0.60,
    ),
    'gpt-4o-mini-realtime-preview': types.ProviderModelPricingType(
      per_response_token_cost=0.60,
      per_query_token_cost=2.40,
    ),
    'o1': types.ProviderModelPricingType(
      per_response_token_cost=15.00,
      per_query_token_cost=60.00,
    ),
    'o1-pro': types.ProviderModelPricingType(
      per_response_token_cost=150.00,
      per_query_token_cost=600.00,
    ),
    'o3': types.ProviderModelPricingType(
      per_response_token_cost=10.00,
      per_query_token_cost=40.00,
    ),
    'o4-mini': types.ProviderModelPricingType(
      per_response_token_cost=1.10,
      per_query_token_cost=4.40,
    ),
    'o3-mini': types.ProviderModelPricingType(
      per_response_token_cost=1.10,
      per_query_token_cost=4.40,
    ),
    'o1-mini': types.ProviderModelPricingType(
      per_response_token_cost=1.10,
      per_query_token_cost=4.40,
    ),
    'gpt-4o-mini-search-preview': types.ProviderModelPricingType(
      per_response_token_cost=0.15,
      per_query_token_cost=0.60,
    ),
    'gpt-4o-search-preview': types.ProviderModelPricingType(
      per_response_token_cost=2.50,
      per_query_token_cost=10.00,
    ),
    'computer-use-preview': types.ProviderModelPricingType(
      per_response_token_cost=3.00,
      per_query_token_cost=12.00,
    ),
    'chatgpt-4o-latest': types.ProviderModelPricingType(
      per_response_token_cost=5.00,
      per_query_token_cost=15.00,
    ),
    'gpt-4-turbo': types.ProviderModelPricingType(
      per_response_token_cost=10.00,
      per_query_token_cost=30.00,
    ),
    'gpt-4': types.ProviderModelPricingType(
      per_response_token_cost=30.00,
      per_query_token_cost=60.00,
    ),
    'gpt-4-32k': types.ProviderModelPricingType(
      per_response_token_cost=60.00,
      per_query_token_cost=120.00,
    ),
    'gpt-3.5-turbo': types.ProviderModelPricingType(
      per_response_token_cost=0.50,
      per_query_token_cost=1.50,
    ),
  }),

  # Claude
  'claude': MappingProxyType({
    'sonnet': types.ProviderModelPricingType(
      per_response_token_cost=3.0,
      per_query_token_cost=15.0,
    ),
    'haiku': types.ProviderModelPricingType(
      per_response_token_cost=0.8,
      per_query_token_cost=4.0,
    ),
    '3.5-sonnet-v2': types.ProviderModelPricingType(
      per_response_token_cost=3.0,
      per_query_token_cost=15.0,
    ),
    '3.5-sonnet': types.ProviderModelPricingType(
      per_response_token_cost=3.0,
      per_query_token_cost=15.0,
    ),
    'opus': types.ProviderModelPricingType(
      per_response_token_cost=15.0,
      per_query_token_cost=75.0,
    ),
    '3-sonnet': types.ProviderModelPricingType(
      per_response_token_cost=3.00,
      per_query_token_cost=15.00,
    ),
    '3-haiku': types.ProviderModelPricingType(
      per_response_token_cost=0.25,
      per_query_token_cost=1.25,
    ),
  }),

  # Gemini
  'gemini': MappingProxyType({
    'gemini-1.5-pro': types.ProviderModelPricingType(
      per_response_token_cost=5.00,
      per_query_token_cost=1.25,
    ),
    'gemini-1.5-flash-8b': types.ProviderModelPricingType(
      per_response_token_cost=0.15,
      per_query_token_cost=0.0375,
    ),
    'gemini-1.5-flash': types.ProviderModelPricingType(
      per_response_token_cost=0.075,
      per_query_token_cost=0.30,
    ),
    'gemini-2.0-flash-lite': types.ProviderModelPricingType(
      per_response_token_cost=0.075,
      per_query_token_cost=0.30,
    ),
    'gemini-2.0-flash': types.ProviderModelPricingType(
      per_response_token_cost=0.1,
      per_query_token_cost=0.4,
    ),
    'gemini-2.5-flash-preview': types.ProviderModelPricingType(
      per_response_token_cost=0.15,
      per_query_token_cost=0.60,
    ),
    'gemini-2.5-pro-preview': types.ProviderModelPricingType(
      per_response_token_cost=1.25,
      per_query_token_cost=10.00,
    ),
  }),

  # Cohere
  'cohere': MappingProxyType({
    'command-a': types.ProviderModelPricingType(
      per_response_token_cost=2.5,
      per_query_token_cost=10.0,
    ),
    'command-r7b': types.ProviderModelPricingType(
      per_response_token_cost=0.0375,
      per_query_token_cost=0.15,
    ),
    'command-r-plus': types.ProviderModelPricingType(
      per_response_token_cost=2.5,
      per_query_token_cost=10.0,
    ),
    'command-r-08-2024': types.ProviderModelPricingType(
      per_response_token_cost=0.15,
      per_query_token_cost=0.6,
    ),
    'command-r': types.ProviderModelPricingType(
      per_response_token_cost=0.15,
      per_query_token_cost=0.6,
    ),
    'command': types.ProviderModelPricingType(
      per_response_token_cost=0.5,
      per_query_token_cost=1.5,
    ),
    'command-nightly': types.ProviderModelPricingType(
      per_response_token_cost=0.5,
      per_query_token_cost=1.5,
    ),
    'command-light': types.ProviderModelPricingType(
      per_response_token_cost=0.0375,
      per_query_token_cost=0.15,
    ),
    'command-light-nightly': types.ProviderModelPricingType(
      per_response_token_cost=0.0375,
      per_query_token_cost=0.15,
    ),
  }),

  # Databricks
  'databricks': MappingProxyType({
    'llama-4-maverick': types.ProviderModelPricingType(
      per_response_token_cost=2.5,
      per_query_token_cost=12.5,
    ),
    'claude-3-7-sonnet': types.ProviderModelPricingType(
      per_response_token_cost=2.5,
      per_query_token_cost=12.5,
    ),
    'meta-llama-3-1-8b-it': types.ProviderModelPricingType(
      per_response_token_cost=2.5,
      per_query_token_cost=12.5,
    ),
    'meta-llama-3-3-70b-it': types.ProviderModelPricingType(
      per_response_token_cost=2.5,
      per_query_token_cost=12.5,
    ),
    # TODO: This is extremely slow model. Until better filtering, it is not
    # included in the list.
    # 'meta-llama-3-1-405b-it': types.ProviderModelPricingType(
    #   per_response_token_cost=2.5,
    #   per_query_token_cost=12.5,
    # ),
    'dbrx-it': types.ProviderModelPricingType(
      per_response_token_cost=2.5,
      per_query_token_cost=12.5,
    ),
    'mixtral-8x7b-it': types.ProviderModelPricingType(
      per_response_token_cost=2.5,
      per_query_token_cost=12.5,
    ),
  }),

  # Mistral
  'mistral': MappingProxyType({
    'codestral': types.ProviderModelPricingType(
      per_response_token_cost=0.2,
      per_query_token_cost=0.6,
    ),
    'ministral-3b': types.ProviderModelPricingType(
      per_response_token_cost=0.04,
      per_query_token_cost=0.04,
    ),
    'ministral-8b': types.ProviderModelPricingType(
      per_response_token_cost=0.1,
      per_query_token_cost=0.1,
    ),
    'mistral-large': types.ProviderModelPricingType(
      per_response_token_cost=2.0,
      per_query_token_cost=6.0,
    ),
    'mistral-medium': types.ProviderModelPricingType(
      per_response_token_cost=0.4,
      per_query_token_cost=2.0,
    ),
    'mistral-saba': types.ProviderModelPricingType(
      per_response_token_cost=0.2,
      per_query_token_cost=0.6,
    ),
    'mistral-small': types.ProviderModelPricingType(
      per_response_token_cost=0.1,
      per_query_token_cost=0.3,
    ),
    'open-mistral-7b': types.ProviderModelPricingType(
      per_response_token_cost=0.25,
      per_query_token_cost=0.25,
    ),
    'open-mistral-nemo': types.ProviderModelPricingType(
      per_response_token_cost=0.15,
      per_query_token_cost=0.15,
    ),
    'open-mixtral-8x22b': types.ProviderModelPricingType(
      per_response_token_cost=2.0,
      per_query_token_cost=6.0,
    ),
    'open-mixtral-8x7b': types.ProviderModelPricingType(
      per_response_token_cost=0.7,
      per_query_token_cost=0.7,
    ),
    'pixtral-12b': types.ProviderModelPricingType(
      per_response_token_cost=0.15,
      per_query_token_cost=0.15,
    ),
    'pixtral-large': types.ProviderModelPricingType(
      per_response_token_cost=2.0,
      per_query_token_cost=6.0,
    ),
  }),

  # Hugging Face
  'huggingface': MappingProxyType({
    'gemma-2-2b-it': types.ProviderModelPricingType(
      per_response_token_cost=1.48,
      per_query_token_cost=3.7,
    ),
    'meta-llama-3.1-8b-it': types.ProviderModelPricingType(
      per_response_token_cost=1.94,
      per_query_token_cost=2.22,
    ),
    'phi-4': types.ProviderModelPricingType(
      per_response_token_cost=33.33,
      per_query_token_cost=37.0,
    ),
    'qwen3-32b': types.ProviderModelPricingType(
      per_response_token_cost=28.57,
      per_query_token_cost=31.74,
    ),
    'deepseek-r1': types.ProviderModelPricingType(
      per_response_token_cost=50.0,
      per_query_token_cost=55.56,
    ),
    'deepseek-v3': types.ProviderModelPricingType(
      per_response_token_cost=25.0,
      per_query_token_cost=27.78,
    ),
  }),

  # DeepSeek
  'deepseek': MappingProxyType({
    'deepseek-v3': types.ProviderModelPricingType(
      per_response_token_cost=0.27,
      per_query_token_cost=1.10,
    ),
    'deepseek-r1': types.ProviderModelPricingType(
      per_response_token_cost=0.55,
      per_query_token_cost=2.19,
    ),
  }),

  # Grok
  'grok': MappingProxyType({
    'grok-3-beta': types.ProviderModelPricingType(
      per_response_token_cost=3.00,
      per_query_token_cost=15.00,
    ),
    'grok-3-fast-beta': types.ProviderModelPricingType(
      per_response_token_cost=5.00,
      per_query_token_cost=25.00,
    ),
    'grok-3-mini-beta': types.ProviderModelPricingType(
      per_response_token_cost=0.30,
      per_query_token_cost=0.50,
    ),
    'grok-3-mini-fast-beta': types.ProviderModelPricingType(
      per_response_token_cost=0.60,
      per_query_token_cost=4.00,
    ),
  }),
})


def get_provider_model_cost(
    provider_model_identifier: types.ProviderModelIdentifierType,
    query_token_count: int,
    response_token_count: int,
) -> int:
  provider_model = model_configs.get_provider_model_config(
      provider_model_identifier)
  pricing = GENERATE_TEXT_PRICING[provider_model.provider][provider_model.model]
  return math.floor(query_token_count * pricing.per_query_token_cost +
                    response_token_count * pricing.per_response_token_cost)



import dataclasses
import datetime
import enum
from typing import Dict, List, Optional, Tuple, Set, Union
from abc import ABC


class RunType(enum.Enum):
  PRODUCTION = 'PRODUCTION'
  TEST = 'TEST'


class CallType(str, enum.Enum):
  GENERATE_TEXT = 'GENERATE_TEXT'


@dataclasses.dataclass(frozen=True)
class ProviderModelType:
  provider: str
  model: str
  provider_model_identifier: str

  def __str__(self):
    return f'({self.provider}, {self.model})'

  def __repr__(self):
    return (
        'ProviderModelType('
        f'provider={self.provider}, '
        f'model={self.model}, '
        f'provider_model_identifier={self.provider_model_identifier})')

  def __lt__(self, other):
    if not isinstance(other, ProviderModelType):
      return NotImplemented
    return str(self) < str(other)

  def __gt__(self, other):
    if not isinstance(other, ProviderModelType):
      return NotImplemented
    return str(self) > str(other)

  def __le__(self, other):
    if not isinstance(other, ProviderModelType):
      return NotImplemented
    return str(self) <= str(other)

  def __ge__(self, other):
    if not isinstance(other, ProviderModelType):
      return NotImplemented
    return str(self) >= str(other)


ProviderModelTupleType = Tuple[str, str]  # (provider, model) without model_signature
ProviderModelIdentifierType = Union[ProviderModelType, ProviderModelTupleType]
StopType = Union[str, List[str]]
MessagesType = List[Dict[str, str]]


class ModelSizeType(str, enum.Enum):
  SMALL = 'small'
  MEDIUM = 'medium'
  LARGE = 'large'
  LARGEST = 'largest'

ModelSizeIdentifierType = Union[ModelSizeType, str]


@dataclasses.dataclass
class ProviderModelPricingType:
  per_response_token_cost: float
  per_query_token_cost: float


@dataclasses.dataclass
class ProviderModelFeatureType:
  not_supported_features: List[str] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class LoggingOptions:
  logging_path: Optional[str] = None
  stdout: bool = False
  hide_sensitive_content: bool = False


class LoggingType(str, enum.Enum):
  QUERY = 'QUERY'
  ERROR = 'ERROR'
  WARNING = 'WARNING'
  INFO = 'INFO'


@dataclasses.dataclass
class CacheOptions:
  cache_path: Optional[str] = None

  unique_response_limit: Optional[int] = 1
  retry_if_error_cached: bool = False
  clear_query_cache_on_connect: bool = False

  disable_model_cache: bool = False
  clear_model_cache_on_connect: bool = False
  model_cache_duration: Optional[int] = None


@dataclasses.dataclass
class ProxDashOptions:
  stdout: bool = False
  hide_sensitive_content: bool = False
  disable_proxdash: bool = False
  api_key: Optional[str] = None
  base_url: Optional[str] = 'https://proxainest-production.up.railway.app'


@dataclasses.dataclass
class SummaryOptions:
  json: bool = True


@dataclasses.dataclass
class RunOptions:
  run_type: Optional[RunType] = None
  hidden_run_key: Optional[str] = None
  experiment_path: Optional[str] = None
  root_logging_path: Optional[str] = None
  default_model_cache_path: Optional[str] = None
  logging_options: Optional[LoggingOptions] = None
  cache_options: Optional[CacheOptions] = None
  proxdash_options: Optional[ProxDashOptions] = None
  allow_multiprocessing: Optional[bool] = None
  model_test_timeout: Optional[int] = None
  strict_feature_test: Optional[bool] = None
  suppress_provider_errors: Optional[bool] = None


@dataclasses.dataclass
class QueryRecord:
  call_type: Optional[CallType] = None
  provider_model: Optional[ProviderModelType] = None
  prompt: Optional[str] = None
  system: Optional[str] = None
  messages: Optional[MessagesType] = None
  max_tokens: Optional[int] = None
  temperature: Optional[float] = None
  stop: Optional[StopType] = None
  hash_value: Optional[str] = None
  token_count: Optional[int] = None


@dataclasses.dataclass
class QueryResponseRecord:
  response: Optional[str] = None
  error: Optional[str] = None
  error_traceback: Optional[str] = None
  start_utc_date: Optional[datetime.datetime] = None
  end_utc_date: Optional[datetime.datetime] = None
  local_time_offset_minute: Optional[int] = None
  response_time: Optional[datetime.timedelta] = None
  estimated_cost: Optional[int] = None
  token_count: Optional[int] = None


@dataclasses.dataclass
class CacheRecord:
  query_record: Optional[QueryRecord] = None
  query_responses: List[QueryResponseRecord] = dataclasses.field(
      default_factory=list)
  shard_id: Optional[str] = None
  last_access_time: Optional[datetime.datetime] = None
  call_count: Optional[int] = None


@dataclasses.dataclass
class LightCacheRecord:
  query_record_hash: Optional[str] = None
  query_response_count: Optional[int] = None
  shard_id: Optional[int] = None
  last_access_time: Optional[datetime.datetime] = None
  call_count: Optional[int] = None


class CacheLookFailReason(str, enum.Enum):
  CACHE_NOT_FOUND = 'CACHE_NOT_FOUND'
  CACHE_NOT_MATCHED = 'CACHE_NOT_MATCHED'
  UNIQUE_RESPONSE_LIMIT_NOT_REACHED = 'UNIQUE_RESPONSE_LIMIT_NOT_REACHED'
  PROVIDER_ERROR_CACHED = 'PROVIDER_ERROR_CACHED'


@dataclasses.dataclass
class CacheLookResult:
  query_response: Optional[QueryResponseRecord] = None
  look_fail_reason: Optional[CacheLookFailReason] = None


class ResponseSource(str, enum.Enum):
  CACHE = 'CACHE'
  PROVIDER = 'PROVIDER'


@dataclasses.dataclass
class LoggingRecord:
  query_record: Optional[QueryRecord] = None
  response_record: Optional[QueryResponseRecord] = None
  response_source: Optional[ResponseSource] = None
  look_fail_reason: Optional[CacheLookFailReason] = None


@dataclasses.dataclass
class ModelStatus:
  unprocessed_models: Set[ProviderModelType] = dataclasses.field(
      default_factory=set)
  working_models: Set[ProviderModelType] = dataclasses.field(
      default_factory=set)
  failed_models: Set[ProviderModelType] = dataclasses.field(
      default_factory=set)
  filtered_models: Set[ProviderModelType] = dataclasses.field(
      default_factory=set)
  provider_queries: Dict[ProviderModelType, LoggingRecord] = (
      dataclasses.field(default_factory=dict))


ModelStatusByCallType = Dict[CallType, ModelStatus]


class ModelCacheManagerStatus(str, enum.Enum):
  INITIALIZING = 'INITIALIZING'
  CACHE_OPTIONS_NOT_FOUND = 'CACHE_OPTIONS_NOT_FOUND'
  CACHE_PATH_NOT_FOUND = 'CACHE_PATH_NOT_FOUND'
  CACHE_PATH_NOT_WRITABLE = 'CACHE_PATH_NOT_WRITABLE'
  DISABLED = 'DISABLED'
  WORKING = 'WORKING'


class QueryCacheManagerStatus(str, enum.Enum):
  INITIALIZING = 'INITIALIZING'
  CACHE_OPTIONS_NOT_FOUND = 'CACHE_OPTIONS_NOT_FOUND'
  CACHE_PATH_NOT_FOUND = 'CACHE_PATH_NOT_FOUND'
  CACHE_PATH_NOT_WRITABLE = 'CACHE_PATH_NOT_WRITABLE'
  DISABLED = 'DISABLED'
  WORKING = 'WORKING'


class ProxDashConnectionStatus(str, enum.Enum):
  INITIALIZING = 'INITIALIZING'
  DISABLED = 'DISABLED'
  API_KEY_NOT_FOUND = 'API_KEY_NOT_FOUND'
  API_KEY_NOT_VALID = 'API_KEY_NOT_VALID'
  PROXDASH_INVALID_RETURN = 'PROXDASH_INVALID_RETURN'
  CONNECTED = 'CONNECTED'


class StateContainer(ABC):
    """Base class for all state objects in the system."""
    pass


@dataclasses.dataclass
class ModelCacheManagerState(StateContainer):
  status: Optional[ModelCacheManagerStatus] = None
  cache_options: Optional[CacheOptions] = None


@dataclasses.dataclass
class QueryCacheManagerState(StateContainer):
  status: Optional[QueryCacheManagerStatus] = None
  cache_options: Optional[CacheOptions] = None
  shard_count: Optional[int] = 800
  response_per_file: Optional[int] = 200
  cache_response_size: Optional[int] = 40000


@dataclasses.dataclass
class ProxDashConnectionState(StateContainer):
  status: Optional[ProxDashConnectionStatus] = None
  hidden_run_key: Optional[str] = None
  experiment_path: Optional[str] = None
  logging_options: Optional[LoggingOptions] = None
  proxdash_options: Optional[ProxDashOptions] = None
  key_info_from_proxdash: Optional[Dict] = None
  connected_experiment_path: Optional[str] = None


@dataclasses.dataclass
class ProviderModelState(StateContainer):
  provider_model: Optional[ProviderModelType] = None
  run_type: Optional[RunType] = None
  strict_feature_test: Optional[bool] = None
  query_cache_manager: Optional[QueryCacheManagerState] = None
  logging_options: Optional[LoggingOptions] = None
  proxdash_connection: Optional[ProxDashConnectionState] = None


@dataclasses.dataclass
class AvailableModelsState(StateContainer):
  run_type: Optional[RunType] = None
  model_cache_manager: Optional[ModelCacheManagerState] = None
  logging_options: Optional[LoggingOptions] = None
  proxdash_connection: Optional[ProxDashConnectionState] = None
  allow_multiprocessing: Optional[bool] = None
  model_test_timeout: Optional[int] = None
  providers_with_key: Optional[Set[str]] = None
  has_fetched_all_models: Optional[bool] = None
  latest_model_cache_path_used_for_update: Optional[str] = None



import proxai.types as types
import proxai.connectors.model_configs as model_configs
from typing import Any


def check_provider_model_identifier_type(
    provider_model_identifier: types.ProviderModelIdentifierType):
  """Check if provider model identifier is supported."""
  if isinstance(provider_model_identifier, types.ProviderModelType):
    provider = provider_model_identifier.provider
    model = provider_model_identifier.model
    if provider not in model_configs.ALL_MODELS:
      raise ValueError(
        f'Provider not supported: {provider}.\n'
        f'Supported providers: {model_configs.ALL_MODELS.keys()}')
    if model not in model_configs.ALL_MODELS[provider]:
      raise ValueError(
        f'Model not supported: {model}.\nSupported models: '
        f'{model_configs.ALL_MODELS[provider].keys()}')
    if provider_model_identifier != model_configs.ALL_MODELS[provider][model]:
      raise ValueError(
        'Mismatch between provider model identifier and model config.'
        f'Provider model identifier: {provider_model_identifier}'
        f'Model config: {model_configs.ALL_MODELS[provider][model]}')
  elif is_provider_model_tuple(provider_model_identifier):
    provider = provider_model_identifier[0]
    model = provider_model_identifier[1]
    if provider not in model_configs.ALL_MODELS:
      raise ValueError(
        f'Provider not supported: {provider}.\n'
        f'Supported providers: {model_configs.ALL_MODELS.keys()}')
    if model not in model_configs.ALL_MODELS[provider]:
      raise ValueError(
        f'Model not supported: {model}.\nSupported models: '
        f'{model_configs.ALL_MODELS[provider].keys()}')
  else:
    raise ValueError(
        f'Invalid provider model identifier: {provider_model_identifier}')


def check_messages_type(messages: types.MessagesType):
  """Check if messages type is supported."""
  for message in messages:
    if not isinstance(message, dict):
      raise ValueError(
          f'Each message in messages should be a dictionary. '
          f'Invalid message: {message}')
    if set(list(message.keys())) != {'role', 'content'}:
      raise ValueError(
          f'Each message should have keys "role" and "content". '
          f'Invalid message: {message}')
    if not isinstance(message['role'], str):
      raise ValueError(
          f'Role should be a string. Invalid role: {message["role"]}')
    if not isinstance(message['content'], str):
      raise ValueError(
          f'Content should be a string. Invalid content: {message["content"]}')
    if message['role'] not in ['user', 'assistant']:
      raise ValueError(
          'Role should be "user" or "assistant".\n'
          f'Invalid role: {message["role"]}')


def is_provider_model_tuple(value: Any) -> bool:
    return (type(value) == tuple
            and len(value) == 2
            and type(value[0]) == str
            and type(value[1]) == str)


def check_model_size_identifier_type(
    model_size_identifier: types.ModelSizeIdentifierType
) -> types.ModelSizeType:
  """Check if model size identifier is supported."""
  if isinstance(model_size_identifier, types.ModelSizeType):
    return model_size_identifier
  elif type(model_size_identifier) == str:
    valid_values = [size.value for size in types.ModelSizeType]
    if model_size_identifier not in valid_values:
      raise ValueError(
          'Model size should be proxai.types.ModelSizeType or one of the '
          'following strings: small, medium, large, largest\n'
          f'Invalid model size identifier: {model_size_identifier}')
    return types.ModelSizeType(model_size_identifier)
  raise ValueError(
        'Model size should be proxai.types.ModelSizeType or one of the '
        'following strings: small, medium, large, largest\n'
        f'Invalid model size identifier: {model_size_identifier}\n'
        f'Type: {type(model_size_identifier)}')



import random
import datetime
import string

MAX_COMPONENT_LENGTH = 64
MAX_PATH_LENGTH = 500
RESERVED_NAMES = {
    'con', 'prn', 'aux', 'nul', 'com1', 'com2', 'com3', 'com4',
    'com5', 'com6', 'com7', 'com8', 'com9', 'lpt1', 'lpt2',
    'lpt3', 'lpt4', 'lpt5', 'lpt6', 'lpt7', 'lpt8', 'lpt9'
}

def get_hidden_run_key() -> str:
  random.seed(datetime.datetime.now().timestamp())
  return str(random.randint(1, 1000000))


def validate_experiment_path(experiment_path: str):
  if not experiment_path:
    raise ValueError('Experiment path cannot be empty')

  if len(experiment_path) > MAX_PATH_LENGTH:
    raise ValueError(
      f'Experiment path cannot be longer than {MAX_PATH_LENGTH} characters')

  allowed_chars = set(string.ascii_letters + string.digits + '()_-.:/ ')
  for char in experiment_path:
    if char not in allowed_chars:
      raise ValueError(
          'Experiment path can only contain following characters:\n'
          f'{sorted(list(allowed_chars))}')
  if experiment_path.startswith('/'):
    raise ValueError('Experiment path cannot start with "/"')
  if experiment_path.endswith('/'):
    raise ValueError('Experiment path cannot end with "/"')
  if '//' in experiment_path:
    raise ValueError('Experiment path cannot contain "//"')
  if '..' in experiment_path:
    raise ValueError('Experiment path cannot contain ".."')
  if '  ' in experiment_path:
    raise ValueError('Experiment path cannot contain consecutive spaces')
  for component in experiment_path.split('/'):
    if len(component) > MAX_COMPONENT_LENGTH:
      raise ValueError(
        f'Path component "{component}" exceeds maximum length of '
        f'{MAX_COMPONENT_LENGTH}')
    if component == '.':
      raise ValueError('Filename in experiment path cannot be "."')
    if component == '-':
      raise ValueError('Filename in experiment path cannot be "-"')
    if component == '_':
      raise ValueError('Filename in experiment path cannot be "_"')
    if component.lower() in RESERVED_NAMES:
      raise ValueError(f'Path component "{component}" is a reserved name')
    if component.startswith(' ') or component.endswith(' '):
      raise ValueError(
          f'Path component "{component}" cannot start or end with spaces')



import copy
import functools
import cohere
import proxai.types as types
import proxai.connectors.providers.cohere_api_mock as cohere_api_mock
import proxai.connectors.model_connector as model_connector


class CohereConnector(model_connector.ProviderModelConnector):
  def get_provider_name(self):
    return 'cohere'

  def init_model(self):
    return cohere.Client()

  def init_mock_model(self):
    return cohere_api_mock.CohereMock()

  def generate_text_proc(self, query_record: types.QueryRecord) -> str:
    # Note: Cohere uses 'SYSTEM', 'USER', and 'CHATBOT' as roles. Additionally,
    # system instructions can be provided in two ways: preamble parameter and
    # chat_history 'SYSTEM' role. The difference is explained in the
    # documentation. The suggested way is to use the preamble parameter.
    query_messages = []
    prompt = query_record.prompt
    if query_record.messages is not None:
      for message in query_record.messages:
        if message['role'] == 'user':
          query_messages.append(
              {'role': 'USER', 'message': message['content']})
        if message['role'] == 'assistant':
          query_messages.append(
              {'role': 'CHATBOT', 'message': message['content']})
      prompt = query_messages[-1]['message']
      del query_messages[-1]
    provider_model = query_record.provider_model

    create = functools.partial(
        self.api.chat,
        model=provider_model.provider_model_identifier,
        message=prompt)
    if query_record.system is not None:
      create = functools.partial(create, preamble=query_record.system)
    if query_messages:
      create = functools.partial(create, chat_history=query_messages)
    if query_record.max_tokens is not None:
      create = functools.partial(create, max_tokens=query_record.max_tokens)
    if query_record.temperature is not None:
      create = functools.partial(create, temperature=query_record.temperature)
    if query_record.stop is not None:
      if isinstance(query_record.stop, list):
        create = functools.partial(create, stop_sequences=query_record.stop)
      else:
        create = functools.partial(create, stop_sequences=[query_record.stop])

    completion = create()
    return completion.text
